import json
from typing import Optional
from unittest import TestCase, mock

from pcs import settings
from pcs.common import file_type_codes, reports
from pcs.common.communication.const import COM_STATUS_ERROR, COM_STATUS_SUCCESS
from pcs.common.communication.dto import InternalCommunicationResultDto
from pcs.common.communication.types import CommunicationResultStatus
from pcs.common.host import Destination, PcsKnownHost
from pcs.common.interface.dto import to_dict
from pcs.common.pcs_cfgsync_dto import SyncConfigsDto
from pcs.common.reports.processor import ReportProcessorToLog
from pcs.daemon.pcs_cfgsync import CfgSyncPullManager
from pcs.lib.corosync.config_facade import ConfigFacade as CorosyncFacade
from pcs.lib.corosync.config_parser import Exporter as CorosyncExporter
from pcs.lib.host.config.exporter import Exporter as KnownHostsExporter
from pcs.lib.host.config.types import KnownHosts
from pcs.lib.permissions.config.exporter import (
    ExporterV2 as PcsSettingsExporter,
)
from pcs.lib.permissions.config.types import (
    ClusterPermissions,
)
from pcs.lib.permissions.config.types import (
    ConfigV2 as PcsSettingsConf,
)

from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import read_test_resource


def fixture_communication_result_string(
    status: CommunicationResultStatus = COM_STATUS_SUCCESS,
    status_msg: Optional[str] = None,
    report_list: Optional[reports.dto.ReportItemDto] = None,
    data="",
) -> str:
    return json.dumps(
        to_dict(
            InternalCommunicationResultDto(
                status=status,
                status_msg=status_msg,
                report_list=report_list or [],
                data=data,
            )
        )
    )


def fixture_known_hosts_content(data_version=1, known_hosts=None):
    return KnownHostsExporter.export(
        KnownHosts(
            format_version=1,
            data_version=data_version,
            known_hosts=known_hosts or {},
        )
    ).decode("utf-8")


def fixture_pcs_settings_content(data_version=1):
    return PcsSettingsExporter.export(
        PcsSettingsConf(
            data_version=data_version,
            clusters=[],
            permissions=ClusterPermissions([]),
        )
    ).decode("utf-8")


def fixture_corosync_with_no_nodes():
    return CorosyncExporter.export(
        CorosyncFacade.create("test", [], "udp").config
    )


def fixture_corosync_3_nodes():
    return read_test_resource("corosync-3nodes.conf").encode("utf-8")


def fixture_known_hosts_3_nodes():
    return fixture_known_hosts_content(
        known_hosts={
            "rh7-1": PcsKnownHost(
                "rh7-1", "TOKEN", [Destination("rh7-1", 2224)]
            ),
            "rh7-2": PcsKnownHost(
                "rh7-2", "TOKEN", [Destination("rh7-2", 2224)]
            ),
            "rh7-3": PcsKnownHost(
                "rh7-3", "TOKEN", [Destination("rh7-3", 2224)]
            ),
        }
    )


class RunCfgSync(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.logger = None

    def run_cfgsync(self):
        env = self.env_assist.get_env()
        self.logger = env.logger
        report_processor = ReportProcessorToLog(self.logger)
        puller = CfgSyncPullManager(
            report_processor, env.get_node_communicator(), self.logger
        )
        return puller.run_cfgsync()

    def assert_logger_calls(self, expected_calls):
        self.assertIsNotNone(
            self.logger,
            "You need to call `self.run_cfgsync()` before calling this assert",
        )
        self.logger.assert_has_calls(expected_calls, any_order=False)
        real_calls = self.logger.mock_calls
        self.assertEqual(
            len(real_calls),
            len(expected_calls),
            msg=(
                f"\nreal calls: {real_calls}\nexpected calls: {expected_calls}"
            ),
        )

    def fixture_before_fetch_config_files_all_successful(self):
        self.config.raw_file.exists(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            name="pcs_cfgsync_ctl.exists",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            content="{}",
            name="pcs_cfgsync_ctl.read",
        )
        self.config.raw_file.exists(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            name="corosync_conf.exists",
        )
        self.config.raw_file.read(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            content=fixture_corosync_3_nodes(),
            name="corosync_conf.read",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_3_nodes(),
            name="known_hosts.read",
        )

    def test_no_corosync_conf(self):
        self.config.raw_file.exists(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            name="pcs_cfgsync_ctl.exists",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            content="{}",
        )
        self.config.raw_file.exists(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            exists=False,
            name="corosync_conf.exists",
        )

        result = self.run_cfgsync()

        self.assert_logger_calls(
            [
                mock.call.info("Config files sync started"),
                mock.call.info(
                    "Config files sync skipped, this host does not seem to be in a cluster of at least 2 nodes"
                ),
            ]
        )
        self.assertEqual(result, settings.pcs_cfgsync_thread_interval_default)

    def test_no_corosync_no_cfgsync_ctl(self):
        self.config.raw_file.exists(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            exists=False,
            name="pcs_cfgsync_ctl.exists",
        )
        self.config.raw_file.exists(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            exists=False,
            name="corosync_conf.exists",
        )

        result = self.run_cfgsync()

        self.assert_logger_calls(
            [
                mock.call.info("Config files sync started"),
                mock.call.debug(
                    "File '%s' does not exist, using default settings for config files sync",
                    settings.pcs_cfgsync_ctl_location,
                ),
                mock.call.info(
                    "Config files sync skipped, this host does not seem to be in a cluster of at least 2 nodes"
                ),
            ]
        )
        self.assertEqual(result, settings.pcs_cfgsync_thread_interval_default)

    def test_no_corosync_error_reading_cfgsync_ctl(self):
        self.config.raw_file.exists(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            name="pcs_cfgsync_ctl.exists",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            exception_msg="Some error",
        )
        self.config.raw_file.exists(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            exists=False,
            name="corosync_conf.exists",
        )

        result = self.run_cfgsync()

        self.assert_logger_calls(
            [
                mock.call.info("Config files sync started"),
                mock.call.error(
                    f"Unable to read Config synchronization configuration '{settings.pcs_cfgsync_ctl_location}': Some error"
                ),
                mock.call.info(
                    "Unable to read '%s', using default settings for config files sync",
                    settings.pcs_cfgsync_ctl_location,
                ),
                mock.call.info(
                    "Config files sync skipped, this host does not seem to be in a cluster of at least 2 nodes"
                ),
            ]
        )
        self.assertEqual(result, settings.pcs_cfgsync_thread_interval_default)

    def test_no_2_nodes_in_corosync_conf(self):
        self.config.raw_file.exists(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            name="pcs_cfgsync_ctl.exists",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            content="{}",
            name="pcs_cfgsync_ctl.read",
        )
        self.config.raw_file.exists(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            name="corosync_conf.exists",
        )
        self.config.raw_file.read(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            content=fixture_corosync_with_no_nodes(),
            name="corosync_conf.read",
        )

        result = self.run_cfgsync()
        self.assert_logger_calls(
            [
                mock.call.info("Config files sync started"),
                mock.call.info(
                    "Config files sync skipped, this host does not seem to be in a cluster of at least 2 nodes"
                ),
            ]
        )
        self.assertEqual(result, settings.pcs_cfgsync_thread_interval_default)

    def test_error_reading_corosync_conf(self):
        self.config.raw_file.exists(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            name="pcs_cfgsync_ctl.exists",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            content="{}",
            name="pcs_cfgsync_ctl.read",
        )
        self.config.raw_file.exists(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            name="corosync_conf.exists",
        )
        self.config.raw_file.read(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            exception_msg="Some error",
            name="corosync_conf.read",
        )

        result = self.run_cfgsync()
        self.assert_logger_calls(
            [
                mock.call.info("Config files sync started"),
                mock.call.error(
                    f"Unable to read Corosync configuration '{settings.corosync_conf_file}': Some error"
                ),
                mock.call.info(
                    "Config files sync skipped, this host does not seem to be in a cluster of at least 2 nodes"
                ),
            ]
        )
        self.assertEqual(result, settings.pcs_cfgsync_thread_interval_default)

    def test_no_known_hosts(self):
        self.config.raw_file.exists(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            name="pcs_cfgsync_ctl.exists",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            content="{}",
            name="pcs_cfgsync_ctl.read",
        )
        self.config.raw_file.exists(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            name="corosync_conf.exists",
        )
        self.config.raw_file.read(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            content=fixture_corosync_3_nodes(),
            name="corosync_conf.read",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            exception_msg="No file",
            name="known_hosts.read",
        )

        result = self.run_cfgsync()
        self.assert_logger_calls(
            [
                mock.call.info("Config files sync started"),
                mock.call.error(
                    f"Unable to read known-hosts '{settings.pcsd_known_hosts_location}': No file"
                ),
                mock.call.error(
                    "Hosts 'rh7-1', 'rh7-2', 'rh7-3' are not known to pcs"
                ),
                mock.call.error("None of hosts is known to pcs."),
                mock.call.info(
                    "Config files sync skipped, unable to find at least 2 request targets"
                ),
            ]
        )
        self.assertEqual(result, settings.pcs_cfgsync_thread_interval_default)

    def test_hosts_from_corosync_not_in_known_hosts(self):
        self.config.raw_file.exists(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            name="pcs_cfgsync_ctl.exists",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_CFGSYNC_CTL,
            settings.pcs_cfgsync_ctl_location,
            content="{}",
            name="pcs_cfgsync_ctl.read",
        )
        self.config.raw_file.exists(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            name="corosync_conf.exists",
        )
        self.config.raw_file.read(
            file_type_codes.COROSYNC_CONF,
            settings.corosync_conf_file,
            content=fixture_corosync_3_nodes(),
            name="corosync_conf.read",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_content(
                known_hosts={
                    "rh7-1": PcsKnownHost(
                        "rh7-1", "foobar", [Destination("10.0.1.2", 22234)]
                    )
                }
            ),
            name="known_hosts.read",
        )

        result = self.run_cfgsync()
        self.assert_logger_calls(
            [
                mock.call.info("Config files sync started"),
                mock.call.error("Hosts 'rh7-2', 'rh7-3' are not known to pcs"),
                mock.call.info(
                    "Config files sync skipped, unable to find at least 2 request targets"
                ),
            ]
        )
        self.assertEqual(result, settings.pcs_cfgsync_thread_interval_default)

    def test_fetch_known_hosts_not_newer_than_local(self):
        self.fixture_before_fetch_config_files_all_successful()
        known_hosts_file_content = fixture_known_hosts_content()
        self.config.http.place_multinode_call(
            "fetch.get_configs",
            communication_list=[
                {
                    "label": "rh7-1",
                    "output": fixture_communication_result_string(
                        data=SyncConfigsDto(
                            cluster_name="test99",
                            configs={
                                file_type_codes.PCS_KNOWN_HOSTS: known_hosts_file_content
                            },
                        )
                    ),
                },
                {
                    "label": "rh7-2",
                    "output": fixture_communication_result_string(
                        data=SyncConfigsDto(
                            cluster_name="test99",
                            configs={
                                file_type_codes.PCS_KNOWN_HOSTS: known_hosts_file_content
                            },
                        )
                    ),
                },
                {
                    "label": "rh7-3",
                    "output": fixture_communication_result_string(
                        data=SyncConfigsDto(
                            cluster_name="test99",
                            configs={
                                file_type_codes.PCS_KNOWN_HOSTS: known_hosts_file_content
                            },
                        )
                    ),
                },
            ],
            action="api/v1/cfgsync-get-configs/v1",
            raw_data=json.dumps({"cluster_name": "test99"}),
        )
        self.config.raw_file.exists(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.fetch.exists",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=known_hosts_file_content,
            name="known_hosts.fetch.read",
        )

        result = self.run_cfgsync()
        self.assert_logger_calls(
            [
                mock.call.info("Config files sync started"),
                mock.call.info(
                    "Fetching config files from nodes: %s",
                    "'rh7-1', 'rh7-2', 'rh7-3'",
                ),
                mock.call.info("Config files sync finished"),
            ]
        )
        self.assertEqual(result, settings.pcs_cfgsync_thread_interval_default)

    def test_fetch_known_hosts_newer_than_local(self):
        self.fixture_before_fetch_config_files_all_successful()
        known_hosts_older = fixture_known_hosts_content()
        known_hosts_new = fixture_known_hosts_content(data_version=99)

        self.config.http.place_multinode_call(
            "fetch.get_configs",
            communication_list=[
                {
                    "label": "rh7-1",
                    "output": fixture_communication_result_string(
                        data=SyncConfigsDto(
                            cluster_name="test99",
                            configs={
                                file_type_codes.PCS_KNOWN_HOSTS: known_hosts_new
                            },
                        )
                    ),
                },
                {
                    "label": "rh7-2",
                    "output": fixture_communication_result_string(
                        data=SyncConfigsDto(
                            cluster_name="test99",
                            configs={
                                file_type_codes.PCS_KNOWN_HOSTS: known_hosts_older
                            },
                        )
                    ),
                },
                {
                    "label": "rh7-3",
                    "output": fixture_communication_result_string(
                        data=SyncConfigsDto(
                            cluster_name="test99",
                            configs={
                                file_type_codes.PCS_KNOWN_HOSTS: known_hosts_older
                            },
                        )
                    ),
                },
            ],
            action="api/v1/cfgsync-get-configs/v1",
            raw_data=json.dumps({"cluster_name": "test99"}),
        )
        self.config.raw_file.exists(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.fetch.exists",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=known_hosts_older,
            name="known_hosts.fetch.read",
        )
        self.config.raw_file.write(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            file_data=known_hosts_new.encode("utf-8"),
            can_overwrite=True,
            name="known_hosts.fetch.write",
        )

        result = self.run_cfgsync()
        self.assert_logger_calls(
            [
                mock.call.info("Config files sync started"),
                mock.call.info(
                    "Fetching config files from nodes: %s",
                    "'rh7-1', 'rh7-2', 'rh7-3'",
                ),
                mock.call.info(
                    "Saving config '%s' version %d to '%s'",
                    "known-hosts",
                    99,
                    settings.pcsd_known_hosts_location,
                ),
                mock.call.info("Config files sync finished"),
            ]
        )
        self.assertEqual(result, settings.pcs_cfgsync_thread_interval_default)

    def test_fetch_hosts_communication_failure(self):
        self.fixture_before_fetch_config_files_all_successful()
        self.config.http.place_multinode_call(
            "fetch.get_configs",
            communication_list=[
                {
                    "label": "rh7-1",
                    "output": fixture_communication_result_string(
                        status=COM_STATUS_ERROR, status_msg="Some error 1"
                    ),
                },
                {
                    "label": "rh7-2",
                    "output": fixture_communication_result_string(
                        status=COM_STATUS_ERROR, status_msg="Some error 2"
                    ),
                },
                {
                    "label": "rh7-3",
                    "output": fixture_communication_result_string(
                        data=SyncConfigsDto(
                            cluster_name="test99",
                            configs={
                                file_type_codes.PCS_KNOWN_HOSTS: fixture_known_hosts_content()
                            },
                        )
                    ),
                },
            ],
            action="api/v1/cfgsync-get-configs/v1",
            raw_data=json.dumps({"cluster_name": "test99"}),
        )
        self.config.raw_file.exists(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="known_hosts.fetch.exists",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_content(),
            name="known_hosts.fetch.read",
        )

        result = self.run_cfgsync()
        self.assert_logger_calls(
            [
                mock.call.info("Config files sync started"),
                mock.call.info(
                    "Fetching config files from nodes: %s",
                    "'rh7-1', 'rh7-2', 'rh7-3'",
                ),
                mock.call.error("rh7-1: Some error 1"),
                mock.call.error("rh7-2: Some error 2"),
                mock.call.info("Config files sync finished"),
            ]
        )
        self.assertEqual(
            result,
            settings.pcs_cfgsync_thread_interval_previous_not_connected_default,
        )

    def test_multiple_files(self):
        self.fixture_before_fetch_config_files_all_successful()
        known_hosts_old = fixture_known_hosts_content()
        known_hosts_new = fixture_known_hosts_content(data_version=99)
        pcs_settings_old = fixture_pcs_settings_content()
        pcs_settings_new = fixture_pcs_settings_content(data_version=99)

        self.config.http.place_multinode_call(
            "fetch.get_configs",
            communication_list=[
                {
                    "label": "rh7-1",
                    "output": fixture_communication_result_string(
                        data=SyncConfigsDto(
                            cluster_name="test99",
                            configs={
                                file_type_codes.PCS_KNOWN_HOSTS: known_hosts_new,
                                file_type_codes.PCS_SETTINGS_CONF: pcs_settings_new,
                            },
                        )
                    ),
                },
                {
                    "label": "rh7-2",
                    "output": fixture_communication_result_string(
                        data=SyncConfigsDto(
                            cluster_name="test99",
                            configs={
                                file_type_codes.PCS_KNOWN_HOSTS: known_hosts_old,
                                file_type_codes.PCS_SETTINGS_CONF: pcs_settings_old,
                            },
                        )
                    ),
                },
                {
                    "label": "rh7-3",
                    "output": fixture_communication_result_string(
                        data=SyncConfigsDto(
                            cluster_name="test99",
                            configs={
                                file_type_codes.PCS_KNOWN_HOSTS: known_hosts_old,
                                file_type_codes.PCS_SETTINGS_CONF: pcs_settings_old,
                            },
                        )
                    ),
                },
            ],
            action="api/v1/cfgsync-get-configs/v1",
            raw_data=json.dumps({"cluster_name": "test99"}),
        )
        self.config.raw_file.exists(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            name="raw_file.exists.known_hosts",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            content=fixture_known_hosts_content(),
            name="raw_file.read.known_hosts",
        )
        self.config.raw_file.exists(
            file_type_codes.PCS_SETTINGS_CONF,
            settings.pcsd_settings_conf_location,
            name="raw_file.exists.pcs_settings",
        )
        self.config.raw_file.read(
            file_type_codes.PCS_SETTINGS_CONF,
            settings.pcsd_settings_conf_location,
            content=fixture_pcs_settings_content(),
            name="raw_file.read.pcs_settings",
        )
        self.config.raw_file.write(
            file_type_codes.PCS_KNOWN_HOSTS,
            settings.pcsd_known_hosts_location,
            file_data=known_hosts_new.encode("utf-8"),
            can_overwrite=True,
            name="known_hosts.fetch.write",
        )
        self.config.raw_file.write(
            file_type_codes.PCS_SETTINGS_CONF,
            settings.pcsd_settings_conf_location,
            file_data=pcs_settings_new.encode("utf-8"),
            can_overwrite=True,
            name="pcs_settings.fetch.write",
        )

        result = self.run_cfgsync()
        self.assert_logger_calls(
            [
                mock.call.info("Config files sync started"),
                mock.call.info(
                    "Fetching config files from nodes: %s",
                    "'rh7-1', 'rh7-2', 'rh7-3'",
                ),
                mock.call.info(
                    "Saving config '%s' version %d to '%s'",
                    "known-hosts",
                    99,
                    settings.pcsd_known_hosts_location,
                ),
                mock.call.info(
                    "Saving config '%s' version %d to '%s'",
                    "pcs configuration",
                    99,
                    settings.pcsd_settings_conf_location,
                ),
                mock.call.info("Config files sync finished"),
            ]
        )
        self.assertEqual(result, settings.pcs_cfgsync_thread_interval_default)



from unittest import TestCase, mock

from pcs import settings
from pcs.common import file_type_codes, reports
from pcs.common.file import RawFileError
from pcs.common.host import PcsKnownHost
from pcs.common.node_communicator import RequestTarget
from pcs.lib.communication.pcs_cfgsync import ConfigInfo, GetConfigsResult
from pcs.lib.file import metadata
from pcs.lib.file.instance import FileInstance
from pcs.lib.host.config.exporter import Exporter as KnownHostsExporter
from pcs.lib.host.config.facade import Facade as KnownHostsFacade
from pcs.lib.host.config.parser import InvalidFileStructureException
from pcs.lib.host.config.types import KnownHosts
from pcs.lib.pcs_cfgsync.fetcher import ConfigFetcher, _find_newest_config
from pcs.lib.permissions.config.exporter import ExporterV2
from pcs.lib.permissions.config.facade import FacadeV2
from pcs.lib.permissions.config.parser import ParserError
from pcs.lib.permissions.config.types import ClusterPermissions, ConfigV2

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal
from pcs_test.tools.custom_mock import MockLibraryReportProcessor


def fixture_known_hosts(data_version=1, known_hosts=None):
    return KnownHosts(
        format_version=1,
        data_version=data_version,
        known_hosts=known_hosts or {},
    )


def fixture_known_hosts_facade(data_version=1, known_hosts=None):
    return KnownHostsFacade(fixture_known_hosts(data_version, known_hosts))


class FindNewestConfig(TestCase):
    def setUp(self):
        self.instance = FileInstance.for_known_hosts()

    def assert_equal_content(self, real, expected):
        self.assertEqual(real.config, expected.config)

    def test_empty(self):
        result = _find_newest_config(self.instance, [])
        self.assertIsNone(result)

    def test_all_the_same(self):
        config_1 = fixture_known_hosts_facade()
        config_2 = fixture_known_hosts_facade()
        result = _find_newest_config(self.instance, [config_1, config_2])

        self.assert_equal_content(result, config_1)

    def test_one_new(self):
        config_older = fixture_known_hosts_facade()
        config_newest = fixture_known_hosts_facade(data_version=99)
        result = _find_newest_config(
            self.instance,
            [config_older, config_newest],
        )

        self.assert_equal_content(result, config_newest)

    def test_same_version_choose_most_common(self):
        config_1_1 = fixture_known_hosts_facade(
            data_version=99, known_hosts={"A": PcsKnownHost("A", "B", [])}
        )
        config_1_2 = fixture_known_hosts_facade(
            data_version=99, known_hosts={"A": PcsKnownHost("A", "B", [])}
        )
        config_2 = fixture_known_hosts_facade(
            data_version=99, known_hosts={"X": PcsKnownHost("X", "Y", [])}
        )
        result = _find_newest_config(
            self.instance, [config_1_1, config_1_2, config_2]
        )

        self.assert_equal_content(result, config_1_1)

    def test_same_version_same_amount_of_the_same_file(self):
        config_1_1 = fixture_known_hosts_facade(
            data_version=99, known_hosts={"A": PcsKnownHost("A", "B", [])}
        )
        config_1_2 = fixture_known_hosts_facade(
            data_version=99, known_hosts={"A": PcsKnownHost("A", "B", [])}
        )
        config_2_1 = fixture_known_hosts_facade(
            data_version=99, known_hosts={"C": PcsKnownHost("C", "D", [])}
        )
        config_2_2 = fixture_known_hosts_facade(
            data_version=99, known_hosts={"C": PcsKnownHost("C", "D", [])}
        )

        # the function should always return the config with the same content,
        # without being dependent on how the input list was sorted
        configs = [
            [config_1_1, config_1_2, config_2_1, config_2_2],
            [config_1_1, config_2_1, config_1_2, config_2_2],
            [config_2_1, config_2_2, config_1_1, config_1_2],
            [config_2_1, config_1_1, config_2_2, config_1_2],
        ]
        for cfg_list in configs:
            with self.subTest(value=cfg_list):
                result = _find_newest_config(
                    self.instance,
                    [config_1_1, config_1_2, config_2_1, config_2_2],
                )
                self.assert_equal_content(result, config_2_1)


def fixture_known_hosts_file_content(data_version=1, known_hosts=None):
    return KnownHostsExporter.export(
        fixture_known_hosts(data_version, known_hosts)
    ).decode("utf-8")


def fixture_pcs_settings_file_content(data_version=1):
    return ExporterV2.export(
        ConfigV2(
            data_version=data_version,
            clusters=[],
            permissions=ClusterPermissions([]),
        )
    ).decode("utf-8")


@mock.patch("pcs.lib.pcs_cfgsync.fetcher.run")
@mock.patch("pcs.lib.pcs_cfgsync.fetcher.FileInstance.raw_file")
class ConfigFetcherTest(TestCase):
    def setUp(self):
        self.report_processor = MockLibraryReportProcessor()
        self.fetcher = ConfigFetcher(None, self.report_processor)

    def test_no_request_targets(
        self, mock_raw_file: mock.Mock, mock_run: mock.Mock
    ):
        mock_run.return_value = GetConfigsResult(False, {})

        configs, was_successful = self.fetcher.fetch("test", [])

        mock_run.assert_called_once()
        mock_raw_file.exists.assert_not_called()
        self.assertEqual(configs, {})
        self.assertFalse(was_successful)

    def test_local_does_not_exist(
        self, mock_raw_file: mock.Mock, mock_run: mock.Mock
    ):
        file_content = fixture_known_hosts_file_content()
        mock_run.return_value = GetConfigsResult(
            True,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE", file_content)
                ]
            },
        )
        mock_raw_file.exists.return_value = False

        configs, was_successful = self.fetcher.fetch(
            "test", [RequestTarget("NODE")]
        )

        mock_run.assert_called_once()
        mock_raw_file.exists.assert_called_once()
        self.assertEqual(len(configs.keys()), 1)
        self.assertEqual(
            KnownHostsExporter.export(
                configs[file_type_codes.PCS_KNOWN_HOSTS].config
            ).decode("utf-8"),
            file_content,
        )
        self.assertTrue(was_successful)
        assert_report_item_list_equal(
            self.report_processor.report_item_list, []
        )

    @mock.patch("pcs.lib.pcs_cfgsync.fetcher.FileInstance.read_to_facade")
    def test_not_newer_than_local(
        self,
        mock_read_to_facade: mock.Mock,
        mock_raw_file: mock.Mock,
        mock_run: mock.Mock,
    ):
        file_content = fixture_known_hosts_file_content()

        mock_run.return_value = GetConfigsResult(
            True,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE", file_content)
                ]
            },
        )
        mock_raw_file.exists.return_value = True
        mock_read_to_facade.return_value = KnownHostsFacade(
            fixture_known_hosts()
        )

        configs, was_successful = self.fetcher.fetch(
            "test", [RequestTarget("NODE")]
        )

        mock_run.assert_called_once()
        mock_raw_file.exists.assert_called_once()
        mock_read_to_facade.assert_called_once()
        self.assertEqual(configs, {})
        self.assertTrue(was_successful)
        assert_report_item_list_equal(
            self.report_processor.report_item_list, []
        )

    @mock.patch("pcs.lib.pcs_cfgsync.fetcher.FileInstance.read_to_facade")
    def test_newer_than_local(
        self,
        mock_read_to_facade: mock.Mock,
        mock_raw_file: mock.Mock,
        mock_run: mock.Mock,
    ):
        file_content = fixture_known_hosts_file_content(data_version=99)

        mock_run.return_value = GetConfigsResult(
            True,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE", file_content)
                ]
            },
        )
        mock_raw_file.exists.return_value = True
        mock_read_to_facade.return_value = KnownHostsFacade(
            fixture_known_hosts()
        )

        configs, was_successful = self.fetcher.fetch(
            "test", [RequestTarget("NODE")]
        )

        mock_run.assert_called_once()
        mock_raw_file.exists.assert_called_once()
        mock_read_to_facade.assert_called_once()
        self.assertEqual(len(configs.keys()), 1)
        self.assertEqual(
            KnownHostsExporter.export(
                configs[file_type_codes.PCS_KNOWN_HOSTS].config
            ).decode("utf-8"),
            file_content,
        )
        self.assertTrue(was_successful)
        assert_report_item_list_equal(
            self.report_processor.report_item_list, []
        )

    @mock.patch("pcs.lib.pcs_cfgsync.fetcher.FileInstance.read_to_facade")
    def test_multiple_nodes_chooses_newest(
        self,
        mock_read_to_facade: mock.Mock,
        mock_raw_file: mock.Mock,
        mock_run: mock.Mock,
    ):
        file_content_older = fixture_known_hosts_file_content()
        file_content_newer = fixture_known_hosts_file_content(data_version=99)

        mock_run.return_value = GetConfigsResult(
            True,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE-1", file_content_older),
                    ConfigInfo("NODE-2", file_content_newer),
                ]
            },
        )
        mock_raw_file.exists.return_value = True
        mock_read_to_facade.return_value = fixture_known_hosts_facade()

        configs, was_successful = self.fetcher.fetch(
            "test",
            [RequestTarget("NODE-1"), RequestTarget("NODE-2")],
        )

        mock_run.assert_called_once()
        mock_raw_file.exists.assert_called_once()
        mock_read_to_facade.assert_called_once()
        self.assertEqual(len(configs), 1)
        self.assertEqual(
            KnownHostsExporter.export(
                configs[file_type_codes.PCS_KNOWN_HOSTS].config
            ).decode("utf-8"),
            file_content_newer,
        )
        self.assertTrue(was_successful)
        assert_report_item_list_equal(
            self.report_processor.report_item_list, []
        )

    @mock.patch("pcs.lib.pcs_cfgsync.fetcher.FileInstance.read_to_facade")
    def test_multiple_files(
        self,
        mock_read_to_facade: mock.Mock,
        mock_raw_file: mock.Mock,
        mock_run: mock.Mock,
    ):
        known_hosts = fixture_known_hosts_file_content(99)
        pcs_settings = fixture_pcs_settings_file_content(99)

        mock_run.return_value = GetConfigsResult(
            True,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE", known_hosts)
                ],
                file_type_codes.PCS_SETTINGS_CONF: [
                    ConfigInfo("NODE", pcs_settings)
                ],
            },
        )
        mock_raw_file.exists.return_value = True
        mock_read_to_facade.side_effect = [
            fixture_known_hosts_facade(),
            FacadeV2(
                ConfigV2(
                    data_version=1,
                    clusters=[],
                    permissions=ClusterPermissions([]),
                )
            ),
        ]

        configs, was_successful = self.fetcher.fetch(
            "test", [RequestTarget("NODE")]
        )

        mock_run.assert_called_once()
        # called twice
        mock_raw_file.exists.assert_has_calls(
            [mock.call.exists(), mock.call.exists()]
        )
        self.assertEqual(len(mock_raw_file.exists.mock_calls), 2)
        mock_read_to_facade.assert_has_calls([mock.call(), mock.call()])
        self.assertEqual(len(mock_read_to_facade.mock_calls), 2)
        self.assertEqual(len(configs), 2)
        self.assertEqual(
            KnownHostsExporter.export(
                configs[file_type_codes.PCS_KNOWN_HOSTS].config
            ).decode("utf-8"),
            known_hosts,
        )
        self.assertEqual(
            ExporterV2.export(
                configs[file_type_codes.PCS_SETTINGS_CONF].config
            ).decode("utf-8"),
            pcs_settings,
        )
        self.assertTrue(was_successful)
        assert_report_item_list_equal(
            self.report_processor.report_item_list, []
        )

    def test_unsupported_filetype(
        self, mock_raw_file: mock.Mock, mock_run: mock.Mock
    ):
        mock_run.return_value = GetConfigsResult(
            True,
            {file_type_codes.PCSD_SSL_CERT: [ConfigInfo("NODE", "")]},
        )
        configs, was_successful = self.fetcher.fetch(
            "test", [RequestTarget("NODE")]
        )

        mock_run.assert_called_once()
        mock_raw_file.exists.assert_not_called()
        self.assertEqual(len(configs.keys()), 0)
        self.assertTrue(was_successful)
        assert_report_item_list_equal(
            self.report_processor.report_item_list, []
        )

    def test_invalid_remote_file_content(
        self, mock_raw_file: mock.Mock, mock_run: mock.Mock
    ):
        mock_run.return_value = GetConfigsResult(
            True,
            {file_type_codes.PCS_KNOWN_HOSTS: [ConfigInfo("NODE", "{}")]},
        )

        configs, was_successful = self.fetcher.fetch(
            "test", [RequestTarget("NODE")]
        )

        mock_run.assert_called_once()
        mock_raw_file.exists.assert_not_called()
        self.assertEqual(configs, {})
        self.assertTrue(was_successful)
        assert_report_item_list_equal(
            self.report_processor.report_item_list,
            [
                fixture.warn(
                    reports.codes.PARSE_ERROR_INVALID_FILE_STRUCTURE,
                    file_type_code=file_type_codes.PCS_KNOWN_HOSTS,
                    file_path=settings.pcsd_known_hosts_location,
                    reason='missing value for field "format_version"',
                    context=reports.dto.ReportItemContextDto(node="NODE"),
                )
            ],
        )

    @mock.patch("pcs.lib.pcs_cfgsync.fetcher.FileInstance.read_to_facade")
    def test_local_raw_file_error(
        self,
        mock_read_to_facade: mock.Mock,
        mock_raw_file: mock.Mock,
        mock_run: mock.Mock,
    ):
        file_content = fixture_known_hosts_file_content(data_version=99)

        mock_run.return_value = GetConfigsResult(
            True,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE", file_content)
                ]
            },
        )
        mock_raw_file.exists.return_value = True
        mock_read_to_facade.side_effect = RawFileError(
            metadata.for_file_type(file_type_codes.PCS_KNOWN_HOSTS, "AAA"),
            RawFileError.ACTION_READ,
            "foo",
        )

        configs, was_successful = self.fetcher.fetch(
            "test", [RequestTarget("NODE")]
        )

        mock_run.assert_called_once()
        mock_raw_file.exists.assert_called_once_with()
        mock_read_to_facade.assert_called_once_with()
        self.assertEqual(len(configs.keys()), 0)
        self.assertTrue(was_successful)
        assert_report_item_list_equal(
            self.report_processor.report_item_list,
            [
                fixture.error(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=file_type_codes.PCS_KNOWN_HOSTS,
                    operation="read",
                    reason="foo",
                    file_path=settings.pcsd_known_hosts_location,
                )
            ],
        )

    @mock.patch("pcs.lib.pcs_cfgsync.fetcher.FileInstance.read_to_facade")
    def test_local_file_parse_error(
        self,
        mock_read_to_facade: mock.Mock,
        mock_raw_file: mock.Mock,
        mock_run: mock.Mock,
    ):
        file_content = fixture_known_hosts_file_content(data_version=99)

        mock_run.return_value = GetConfigsResult(
            True,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE", file_content)
                ]
            },
        )
        mock_raw_file.exists.return_value = True
        mock_read_to_facade.side_effect = InvalidFileStructureException("AAA")

        configs, was_successful = self.fetcher.fetch(
            "test", [RequestTarget("NODE")]
        )

        mock_run.assert_called_once()
        mock_raw_file.exists.assert_called_once_with()
        mock_read_to_facade.assert_called_once_with()
        self.assertEqual(len(configs.keys()), 0)
        self.assertTrue(was_successful)
        assert_report_item_list_equal(
            self.report_processor.report_item_list,
            [
                fixture.error(
                    reports.codes.PARSE_ERROR_INVALID_FILE_STRUCTURE,
                    reason="AAA",
                    file_type_code=file_type_codes.PCS_KNOWN_HOSTS,
                    file_path=settings.pcsd_known_hosts_location,
                )
            ],
        )

    @mock.patch("pcs.lib.pcs_cfgsync.fetcher.FileInstance.read_to_facade")
    def test_multiple_files_skip_error_local(
        self,
        mock_read_to_facade: mock.Mock,
        mock_raw_file: mock.Mock,
        mock_run: mock.Mock,
    ):
        known_hosts = fixture_known_hosts_file_content(data_version=99)
        pcs_settings = fixture_pcs_settings_file_content(data_version=99)

        mock_run.return_value = GetConfigsResult(
            True,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE", known_hosts)
                ],
                file_type_codes.PCS_SETTINGS_CONF: [
                    ConfigInfo("NODE", pcs_settings)
                ],
            },
        )
        mock_raw_file.exists.return_value = True
        mock_read_to_facade.side_effect = [
            fixture_known_hosts_facade(),
            ParserError("AAA"),
        ]

        configs, was_successful = self.fetcher.fetch(
            "test", [RequestTarget("NODE")]
        )

        mock_run.assert_called_once()
        # called twice
        mock_raw_file.exists.assert_has_calls(
            [mock.call.exists(), mock.call.exists()]
        )
        self.assertEqual(len(mock_raw_file.exists.mock_calls), 2)
        mock_read_to_facade.assert_has_calls([mock.call(), mock.call()])
        self.assertEqual(len(mock_read_to_facade.mock_calls), 2)
        self.assertEqual(len(configs), 1)
        self.assertEqual(
            KnownHostsExporter.export(
                configs[file_type_codes.PCS_KNOWN_HOSTS].config
            ).decode("utf-8"),
            known_hosts,
        )
        self.assertTrue(was_successful)
        assert_report_item_list_equal(
            self.report_processor.report_item_list,
            [
                fixture.error(
                    reports.codes.PARSE_ERROR_INVALID_FILE_STRUCTURE,
                    reason="AAA",
                    file_type_code=file_type_codes.PCS_SETTINGS_CONF,
                    file_path=settings.pcsd_settings_conf_location,
                )
            ],
        )

    @mock.patch("pcs.lib.pcs_cfgsync.fetcher.FileInstance.read_to_facade")
    def test_fetch_filtered(
        self,
        mock_read_to_facade: mock.Mock,
        mock_raw_file: mock.Mock,
        mock_run: mock.Mock,
    ):
        known_hosts = fixture_known_hosts_file_content(99)

        mock_run.return_value = GetConfigsResult(
            True,
            {
                file_type_codes.PCS_KNOWN_HOSTS: [
                    ConfigInfo("NODE", known_hosts)
                ],
            },
        )
        mock_raw_file.exists.return_value = True
        mock_read_to_facade.side_effect = [fixture_known_hosts_facade()]

        configs, was_successful = self.fetcher.fetch(
            "test", [RequestTarget("NODE")], [file_type_codes.PCS_KNOWN_HOSTS]
        )

        mock_run.assert_called_once()
        mock_raw_file.exists.assert_called_once_with()
        mock_read_to_facade.assert_called_once_with()
        self.assertEqual(len(configs), 1)
        self.assertEqual(
            KnownHostsExporter.export(
                configs[file_type_codes.PCS_KNOWN_HOSTS].config
            ).decode("utf-8"),
            known_hosts,
        )
        self.assertTrue(was_successful)
        assert_report_item_list_equal(
            self.report_processor.report_item_list, []
        )



from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import (
    get_test_resource,
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


class QueryTestBase(TestCase, AssertPcsMixin):
    cib = get_test_resource("cib-all.xml")

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_status_query_resource")
        write_file_to_tmpfile(self.cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()


class QueryExists(QueryTestBase):
    def test_true(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "S2", "exists"], "True\n"
        )

    def test_false(self):
        self.assert_pcs_result(
            ["status", "query", "resource", "nonexistent", "exists"],
            stdout_full="False\n",
            stderr_full="",
            returncode=2,
        )


class QueryIsStonith(QueryTestBase):
    def test_true(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "S2", "is-stonith"], "True\n"
        )

    def test_false(self):
        self.assert_pcs_result(
            ["status", "query", "resource", "R7", "is-stonith"],
            stdout_full="False\n",
            stderr_full="",
            returncode=2,
        )

    def test_fail(self):
        self.assert_pcs_fail(
            ["status", "query", "resource", "nonexistent", "is-stonith"],
            "Error: Resource 'nonexistent' does not exist\n",
        )


class QueryIsType(QueryTestBase):
    def test_true(self):
        self.assert_pcs_success(
            [
                "status",
                "query",
                "resource",
                "B1",
                "is-type",
                "bundle",
                "unique",
            ],
            "True\n",
        )

    def test_false(self):
        self.assert_pcs_result(
            ["status", "query", "resource", "G2", "is-type", "clone"],
            stdout_full="False\n",
            stderr_full="",
            returncode=2,
        )


class QueryGetType(QueryTestBase):
    def test_ok(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "G1-clone", "get-type"],
            "clone promotable\n",
        )


class QueryIsInGroup(QueryTestBase):
    def test_group(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "R2", "is-in-group"], "True\nG1\n"
        )

    def test_bad_group(self):
        self.assert_pcs_result(
            ["status", "query", "resource", "R5", "is-in-group", "G1"],
            stdout_full="False\nG2\n",
            stderr_full="",
            returncode=2,
        )

    def test_fail(self):
        self.assert_pcs_fail(
            ["status", "query", "resource", "G1", "is-in-group"],
            (
                "Error: Resource 'G1' has unexpected type 'group'. This "
                "command works only for resources of type 'primitive'\n"
            ),
        )


class QueryIsInClone(QueryTestBase):
    def test_clone(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "R2", "is-in-clone"],
            "True\nG1-clone\n",
        )

    def test_bad_clone(self):
        self.assert_pcs_result(
            ["status", "query", "resource", "R2", "is-in-clone", "R6-clone"],
            stdout_full="False\nG1-clone\n",
            stderr_full="",
            returncode=2,
        )

    def test_fail(self):
        self.assert_pcs_fail(
            ["status", "query", "resource", "R6-clone", "is-in-clone"],
            (
                "Error: Resource 'R6-clone' has unexpected type 'clone'. This "
                "command works only for resources of type 'group', "
                "'primitive'\n"
            ),
        )


class QueryIsInBundle(QueryTestBase):
    def test_in_bundle(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "R1", "is-in-bundle"], "True\nB2\n"
        )

    def test_bad_bundle(self):
        self.assert_pcs_result(
            ["status", "query", "resource", "R1", "is-in-bundle", "B1"],
            stdout_full="False\nB2\n",
            stderr_full="",
            returncode=2,
        )

    def test_fail(self):
        self.assert_pcs_fail(
            ["status", "query", "resource", "G1", "is-in-group"],
            (
                "Error: Resource 'G1' has unexpected type 'group'. This "
                "command works only for resources of type 'primitive'\n"
            ),
        )


class QueryGetMembers(QueryTestBase):
    def test_ok(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "G2", "get-members"],
            "R5\n",
        )

    def test_primitive(self):
        self.assert_pcs_fail(
            ["status", "query", "resource", "S1", "get-members"],
            (
                "Error: Resource 'S1' has unexpected type 'primitive'. This "
                "command works only for resources of type 'bundle', 'clone', "
                "'group'\n"
            ),
        )


class QueryGetIndexInGroup(QueryTestBase):
    def test_ok(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "R5", "get-index-in-group"], "0\n"
        )

    def test_not_in_group(self):
        self.assert_pcs_fail(
            ["status", "query", "resource", "S2", "get-index-in-group"],
            "Error: Resource 'S2' is not in a group\n",
        )


class QueryCibWithStatusTestBase(TestCase, AssertPcsMixin):
    cib = get_test_resource("cib-status.xml")

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_status_query_resource_status")
        write_file_to_tmpfile(self.cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()


class QueryGetNodes(QueryCibWithStatusTestBase):
    def test_one(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "s1", "get-nodes"], "rh93-1\n"
        )

    def test_multiple(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "g1", "get-nodes"],
            "rh93-1\nrh93-2\n",
        )

    def test_fail(self):
        self.assert_pcs_fail(
            ["status", "query", "resource", "g2", "get-nodes"],
            "Error: Resource 'g2' does not exist\n",
        )


class QueryIsState(QueryCibWithStatusTestBase):
    def test_true(self):
        self.assert_pcs_success(
            [
                "status",
                "query",
                "resource",
                "g1",
                "is-state",
                "started",
                "on-node",
                "rh93-1",
            ],
            "True\n",
        )

    def test_false(self):
        self.assert_pcs_result(
            [
                "status",
                "query",
                "resource",
                "g1",
                "is-state",
                "started",
                "members",
                "all",
                "instances",
                "all",
            ],
            stdout_full="False\n",
            stderr_full="",
            returncode=2,
        )

    def test_fail(self):
        self.assert_pcs_fail(
            [
                "status",
                "query",
                "resource",
                "r1",
                "is-state",
                "started",
                "members",
                "all",
            ],
            (
                "Error: 'members' quantifier can be used only on group "
                "resources or group instances of cloned groups\n"
            ),
        )


class QueryResourceRemoteNode(TestCase, AssertPcsMixin):
    cib = get_test_resource("cib-remote.xml")

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_status_query_resource_remote")
        write_file_to_tmpfile(self.cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()

    def test_exists(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "dummy", "exists"], "True\n"
        )

    def test_get_nodes(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "dummy", "get-nodes"],
            "rh93-remote\n",
        )

    def test_started(self):
        self.assert_pcs_success(
            ["status", "query", "resource", "dummy", "is-state", "started"],
            "True\n",
        )



from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import get_test_resource
from pcs_test.tools.pcs_runner import PcsRunner


class ClusterRename(AssertPcsMixin, TestCase):
    def setUp(self):
        self.pcs_runner = PcsRunner(None)

    def test_no_args(self):
        self.assert_pcs_fail(
            ["cluster", "rename"],
            stderr_start="\nUsage: pcs cluster rename...\n",
        )

    def test_too_many_args(self):
        self.assert_pcs_fail(
            ["cluster", "rename", "A", "B"],
            stderr_start="\nUsage: pcs cluster rename...\n",
        )

    def test_not_live_pcmk(self):
        self.pcs_runner = PcsRunner(cib_file=get_test_resource("cib-empty.xml"))
        self.assert_pcs_fail(
            ["cluster", "rename", "A"],
            stderr_full=(
                "Error: Specified option '-f' is not supported in this "
                "command\n"
            ),
        )

    def test_not_live_corosync(self):
        self.pcs_runner = PcsRunner(
            cib_file=None,
            corosync_conf_opt=get_test_resource("corosync_conf"),
        )
        self.assert_pcs_fail(
            ["cluster", "rename", "A"],
            stderr_full=(
                "Error: Specified option '--corosync_conf' is not supported in "
                "this command\n"
            ),
        )



from unittest import TestCase

from pcs.common import validate


class IsInteger(TestCase):
    def test_no_range(self):
        self.assertTrue(validate.is_integer(1))
        self.assertTrue(validate.is_integer("1"))
        self.assertTrue(validate.is_integer(-1))
        self.assertTrue(validate.is_integer("-1"))
        self.assertTrue(validate.is_integer(+1))
        self.assertTrue(validate.is_integer("+1"))

        self.assertFalse(validate.is_integer(" 1"))
        self.assertFalse(validate.is_integer("\n-1"))
        self.assertFalse(validate.is_integer("\r+1"))
        self.assertFalse(validate.is_integer("1\n"))
        self.assertFalse(validate.is_integer("-1 "))
        self.assertFalse(validate.is_integer("+1\r"))

        self.assertFalse(validate.is_integer(""))
        self.assertFalse(validate.is_integer("1a"))
        self.assertFalse(validate.is_integer("a1"))
        self.assertFalse(validate.is_integer("aaa"))
        self.assertFalse(validate.is_integer(1.0))
        self.assertFalse(validate.is_integer("1.0"))

    def test_at_least(self):
        self.assertTrue(validate.is_integer(5, 5))
        self.assertTrue(validate.is_integer(5, 4))
        self.assertTrue(validate.is_integer("5", 5))
        self.assertTrue(validate.is_integer("5", 4))

        self.assertFalse(validate.is_integer(5, 6))
        self.assertFalse(validate.is_integer("5", 6))

    def test_at_most(self):
        self.assertTrue(validate.is_integer(5, None, 5))
        self.assertTrue(validate.is_integer(5, None, 6))
        self.assertTrue(validate.is_integer("5", None, 5))
        self.assertTrue(validate.is_integer("5", None, 6))

        self.assertFalse(validate.is_integer(5, None, 4))
        self.assertFalse(validate.is_integer("5", None, 4))

    def test_range(self):
        self.assertTrue(validate.is_integer(5, 5, 5))
        self.assertTrue(validate.is_integer(5, 4, 6))
        self.assertTrue(validate.is_integer("5", 5, 5))
        self.assertTrue(validate.is_integer("5", 4, 6))

        self.assertFalse(validate.is_integer(3, 4, 6))
        self.assertFalse(validate.is_integer(7, 4, 6))
        self.assertFalse(validate.is_integer("3", 4, 6))
        self.assertFalse(validate.is_integer("7", 4, 6))


class IsPortNumber(TestCase):
    def test_valid_port(self):
        self.assertTrue(validate.is_port_number(1))
        self.assertTrue(validate.is_port_number("1"))
        self.assertTrue(validate.is_port_number(65535))
        self.assertTrue(validate.is_port_number("65535"))
        self.assertTrue(validate.is_port_number(8192))

    def test_bad_port(self):
        self.assertFalse(validate.is_port_number(0))
        self.assertFalse(validate.is_port_number("0"))
        self.assertFalse(validate.is_port_number(65536))
        self.assertFalse(validate.is_port_number("65536"))
        self.assertFalse(validate.is_port_number(" 8192 "))
        self.assertFalse(validate.is_port_number(-128))
        self.assertFalse(validate.is_port_number("-128"))
        self.assertFalse(validate.is_port_number("abcd"))



from unittest import TestCase

from lxml import etree

from pcs.common.str_tools import format_list, format_optional, format_plural
from pcs.lib.resource_agent import const as ra_const

from pcs_test.tools.cib import get_assert_pcs_effect_mixin
from pcs_test.tools.metadata_dto import (
    FIXTURE_KNOWN_META_NAMES_PRIMITIVE_META,
    FIXTURE_KNOWN_META_NAMES_STONITH_META,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_file,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner


def get_cib_resources(cib):
    return etree.tostring(etree.parse(cib).findall(".//resources")[0])


class ResourceTest(TestCase, get_assert_pcs_effect_mixin(get_cib_resources)):
    empty_cib = rc("cib-empty.xml")

    def setUp(self):
        self.temp_cib = get_tmp_file("tier1_test_resource_common")
        write_file_to_tmpfile(self.empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        self.temp_cib.close()


def fixture_meta_attributes_warning(meta_attrs, agent_type):
    type_to_name_list = {
        ra_const.PRIMITIVE_META: FIXTURE_KNOWN_META_NAMES_PRIMITIVE_META,
        ra_const.STONITH_META: FIXTURE_KNOWN_META_NAMES_STONITH_META,
    }
    type_to_rsc_desc = {
        ra_const.PRIMITIVE_META: "resource",
        ra_const.STONITH_META: "stonith",
    }
    if agent_type not in type_to_name_list:
        raise AssertionError(
            f"Unknown agent type '{agent_type}', known types: "
            f"{', '.join(type_to_name_list.keys())}"
        )

    rsc_desc = type_to_rsc_desc[agent_type].capitalize()
    attributes = format_plural(meta_attrs, "attribute")
    have = format_plural(meta_attrs, "has")
    known_meta = format_list(type_to_name_list[agent_type])
    attributes_known = format_plural(type_to_name_list[agent_type], "attribute")
    return (
        f"Warning: {rsc_desc} meta {attributes} {format_list(meta_attrs)} "
        f"{have} no effect on cluster resource handling, meta "
        f"{attributes_known} with effect: {known_meta}\n"
    )


def fixture_meta_attributes_not_validated_warning(meta_type_list):
    resource_desc = format_optional(
        " / ".join(sorted(meta_type_list)), template="of {} "
    )
    return f"Warning: Meta attributes {resource_desc}are not validated\n"


def fixture_use_meta_command_instead_warning(is_stonith=False):
    command = "stonith" if is_stonith else "resource"
    return (
        "Warning: Meta attributes are not validated by this command. For "
        f"validation, please use 'pcs {command} meta' instead.\n"
    )



from logging import Logger
from typing import TYPE_CHECKING, Optional, cast

from pcs.common.node_communicator import Communicator, RequestTarget
from pcs.common.reports.processor import ReportProcessor
from pcs.common.reports.utils import format_file_role
from pcs.common.str_tools import format_list
from pcs.common.types import StringIterable
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.raw_file import RawFileError, raw_file_error_report
from pcs.lib.interface.config import ParserErrorException
from pcs.lib.node_communication import NodeTargetLibFactory
from pcs.lib.pcs_cfgsync.config.facade import Facade as CfgsyncCtlFacade
from pcs.lib.pcs_cfgsync.fetcher import ConfigFetcher

if TYPE_CHECKING:
    from pcs.lib.corosync.config_facade import ConfigFacade as CorosyncFacade
    from pcs.lib.host.config.facade import Facade as KnownHostsFacade


class CfgSyncPullManager:
    def __init__(
        self,
        report_processor: ReportProcessor,
        node_communicator: Communicator,
        logger: Logger,
    ):
        self._logger = logger
        self._report_processor = report_processor
        self._node_communicator = node_communicator
        self._fetcher = ConfigFetcher(
            self._node_communicator, self._report_processor
        )

    def run_cfgsync(self) -> int:
        self._logger.info("Config files sync started")

        ctl_facade = self._read_cfgsync_ctl()

        if not ctl_facade.is_sync_allowed:
            self._logger.info(
                "Config files sync is disabled or paused, skipping"
            )
            return ctl_facade.sync_interval

        cluster_name, node_names = self._get_cluster_info()
        if not cluster_name or len(node_names) < 2:
            self._logger.info(
                "Config files sync skipped, this host does not seem to be in a "
                "cluster of at least 2 nodes"
            )
            return ctl_facade.sync_interval

        target_list = self._get_request_targets(node_names)
        if len(target_list) < 2:
            self._logger.info(
                "Config files sync skipped, unable to find at least 2 request "
                "targets"
            )
            return ctl_facade.sync_interval

        self._logger.info(
            "Fetching config files from nodes: %s",
            format_list(t.label for t in target_list),
        )
        configs, was_connected = self._fetcher.fetch(cluster_name, target_list)
        for file_code, facade in configs.items():
            instance = FileInstance.for_common(file_code)
            try:
                self._logger.info(
                    "Saving config '%s' version %d to '%s'",
                    format_file_role(instance.raw_file.metadata.file_type_code),
                    facade.data_version,
                    instance.raw_file.metadata.path,
                )
                instance.write_facade(facade, can_overwrite=True)
            except RawFileError as e:
                self._report_processor.report(raw_file_error_report(e))

        self._logger.info("Config files sync finished")
        if was_connected:
            return ctl_facade.sync_interval
        return ctl_facade.sync_interval_previous_not_connected

    def _read_cfgsync_ctl(self) -> CfgsyncCtlFacade:
        ctl_instance = FileInstance.for_pcs_cfgsync_ctl()
        if not ctl_instance.raw_file.exists():
            self._logger.debug(
                "File '%s' does not exist, using default settings for config "
                "files sync",
                ctl_instance.raw_file.metadata.path,
            )
            return CfgsyncCtlFacade.create()

        ctl_facade = self._read_file(ctl_instance)
        if ctl_facade is not None:
            return ctl_facade
        self._logger.info(
            "Unable to read '%s', using default settings for config files sync",
            ctl_instance.raw_file.metadata.path,
        )
        return CfgsyncCtlFacade.create()

    def _get_cluster_info(self) -> tuple[str, list[str]]:
        corosync_conf_instance = FileInstance.for_corosync_conf()
        if not corosync_conf_instance.raw_file.exists():
            return "", []
        corosync_conf: Optional[CorosyncFacade] = self._read_file(
            corosync_conf_instance
        )
        if corosync_conf is None:
            return "", []

        cluster_name = corosync_conf.get_cluster_name()
        node_names = [
            node.name
            for node in corosync_conf.get_nodes()
            if node.name is not None
        ]
        return cluster_name, node_names

    def _get_request_targets(
        self, node_names: StringIterable
    ) -> list[RequestTarget]:
        known_hosts_instance = FileInstance.for_known_hosts()
        known_hosts_conf: Optional[KnownHostsFacade] = self._read_file(
            known_hosts_instance
        )

        reports, target_list = NodeTargetLibFactory(
            known_hosts_conf.known_hosts
            if known_hosts_conf is not None
            else {},
            self._report_processor,
        ).get_target_list_with_reports(node_names)
        self._report_processor.report_list(reports)
        return target_list

    def _read_file[T](self, file_instance: FileInstance) -> Optional[T]:
        try:
            return cast(T, file_instance.read_to_facade())
        except RawFileError as e:
            self._report_processor.report(raw_file_error_report(e))
        except ParserErrorException as e:
            self._report_processor.report_list(
                file_instance.parser_exception_to_report_list(e)
            )
        return None



from pcs.cli.reports.processor import print_report
from pcs.common import reports


class NodeRemoveRemoteReportProcessor(reports.ReportProcessor):
    """
    Catch reports in memory without printing them. When we receive the first
    report about node communication, we print all of the saved reports into
    console and then start printing all of the subsequent reports

    Used when we need to catch validation error reports from
    `remote_node.node_remove_remote` library command, but we need to provide
    output to the user after the validations passed but the command has not
    finished yet. When we receive the first report about node communication, we
    can start printing the reports to console since we know that the validations
    have passed and the command started the removal.
    """

    def __init__(self, include_debug: bool = False):
        super().__init__()
        self._include_debug = include_debug
        self._save_in_memory = True
        self._reports: reports.ReportItemList = []

    def _do_report(self, report_item: reports.ReportItem) -> None:
        if (
            report_item.severity.level == reports.ReportItemSeverity.DEBUG
            and not self._include_debug
        ):
            return

        if not self._save_in_memory:
            print_report(report_item.to_dto())
            return

        self._reports.append(report_item)
        if (
            report_item.message.code
            == reports.codes.SERVICE_COMMANDS_ON_NODES_STARTED
        ):
            for saved_report in self._reports:
                print_report(saved_report.to_dto())
            self._save_in_memory = False
            self._reports = []

    @property
    def reports(self) -> reports.ReportItemList:
        return self._reports

    @property
    def already_reported_to_console(self) -> bool:
        return not self._save_in_memory



from typing import Any
from urllib.parse import urlparse

from pcs import (
    settings,
    utils,
)
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import (
    Argv,
    InputModifiers,
    KeyValueParser,
    ensure_unique_args,
    split_list_by_any_keywords,
)
from pcs.common.auth import HostAuthData, HostWithTokenAuthData
from pcs.common.host import Destination
from pcs.common.str_tools import format_list


def _parse_host_options(host: str, options: Argv) -> Destination:
    ADDR_OPT_KEYWORD = "addr"  # pylint: disable=invalid-name
    supported_options = {ADDR_OPT_KEYWORD}
    parsed_options = KeyValueParser(options).get_unique()
    unknown_options = set(parsed_options.keys()) - supported_options
    if unknown_options:
        raise CmdLineInputError(
            f"Unknown options {format_list(unknown_options)} for host '{host}'"
        )
    return _parse_addr(parsed_options.get(ADDR_OPT_KEYWORD, host))


def _parse_addr(addr: str) -> Destination:
    if addr.count(":") > 1 and not addr.startswith("["):
        # if IPv6 without port put it in parentheses
        addr = f"[{addr}]"
    # adding protocol so urlparse will parse hostname/ip and port correctly
    url = urlparse(f"http://{addr}")

    common_exception = CmdLineInputError(
        f"Invalid port number in address '{addr}', use 1..65535"
    )
    # Reading the port attribute will raise a ValueError if an invalid port is
    # specified in the URL.
    try:
        port = url.port
    except ValueError:
        raise common_exception from None
    # urlparse allow 0 as valid port number, pcs does not
    if port == 0:
        raise common_exception
    return Destination(
        addr=url.hostname or "",
        port=port if port else settings.pcsd_default_port,
    )


def auth_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * -u - username
      * -p - password
      * --token - auth token
      * --request-timeout - timeout for HTTP requests
    """
    modifiers.ensure_only_supported("-u", "-p", "--request-timeout", "--token")
    if not argv:
        raise CmdLineInputError("No host specified")
    host_dict = {
        host: [_parse_host_options(host, opts)]
        for host, opts in split_list_by_any_keywords(argv, "host name").items()
    }
    token = modifiers.get("--token")
    if token:
        token_value = utils.get_token_from_file(str(token))
        lib.auth.auth_hosts_token_no_sync(
            {
                host_name: HostWithTokenAuthData(token_value, dest_list)
                for host_name, dest_list in host_dict.items()
            }
        )
        return

    username, password = utils.get_user_and_pass()
    lib.auth.auth_hosts(
        {
            host_name: HostAuthData(username, password, dest_list)
            for host_name, dest_list in host_dict.items()
        }
    )


def deauth_cmd(lib: Any, argv: Argv, modifiers: InputModifiers) -> None:
    """
    Options:
      * --request-timeout - timeout for HTTP requests
    """
    modifiers.ensure_only_supported("--request-timeout")
    ensure_unique_args(argv)

    if not argv:
        lib.auth.deauth_all_local_hosts()
    else:
        lib.auth.deauth_hosts(argv)



# pylint: disable=too-many-lines
import os
from textwrap import dedent
from typing import Optional
from unittest import TestCase, mock

from pcs import settings
from pcs.common import file_type_codes
from pcs.common.const import (
    PCMK_ROLE_STOPPED,
    PCMK_STATUS_ROLE_STOPPED,
    PcmkRoleType,
)
from pcs.common.reports import codes as report_codes
from pcs.common.status_dto import (
    BundleReplicaStatusDto,
    BundleStatusDto,
    CloneStatusDto,
    GroupStatusDto,
    PrimitiveStatusDto,
    ResourcesStatusDto,
)
from pcs.lib.booth import constants
from pcs.lib.commands import status
from pcs.lib.errors import LibraryError

from pcs_test.tools import fixture, fixture_crm_mon
from pcs_test.tools.assertions import assert_xml_equal
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.command_env.config_runner_pcmk import (
    RULE_EXPIRED_RETURNCODE,
    RULE_IN_EFFECT_RETURNCODE,
)
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import read_test_resource as rc_read

EXITCODE_INVALID_CIB = 78


def _booth_config_path_fixture(instance_name="booth"):
    return os.path.join(settings.booth_config_dir, f"{instance_name}.conf")


@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_rng/api/api-result.rng")
)
class PacemakerStatusXml(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_success(self):
        self.config.runner.pcmk.load_state(filename="crm_mon.minimal.xml")
        assert_xml_equal(
            status.pacemaker_status_xml(self.env_assist.get_env()),
            rc_read("crm_mon.minimal.xml"),
        )

    def test_not_xml(self):
        # Loading pacemaker state succeeded, we expect to get an xml and we
        # just return it. If it is not a valid xml, it is a bug in pacemaker.
        self.config.runner.pcmk.load_state(stdout="not an xml")
        self.assertEqual(
            status.pacemaker_status_xml(self.env_assist.get_env()), "not an xml"
        )

    def test_error(self):
        error_xml = fixture_crm_mon.error_xml(
            1, "an error", ["This is an error message", "And one more"]
        )
        self.config.runner.pcmk.load_state(stdout=error_xml, returncode=1)
        with self.assertRaises(LibraryError) as cm:
            status.pacemaker_status_xml(self.env_assist.get_env())
        assert_xml_equal(cm.exception.output, error_xml)

    def test_error_not_xml(self):
        # Loading pacemaker state failed, but we expect to get an xml anyway
        # and we return it. If it is not a valid xml, it is a bug or a critical
        # error in pacemaker we can do nothing about.
        self.config.runner.pcmk.load_state(stdout="an error", returncode=1)
        with self.assertRaises(LibraryError) as cm:
            status.pacemaker_status_xml(self.env_assist.get_env())
        self.assertEqual(cm.exception.output, "an error")


class FullClusterStatusPlaintextBase(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self.node_name_list = ["node1", "node2", "node3"]
        self.maxDiff = None

    @staticmethod
    def _fixture_xml_clustername(name):
        return """
            <crm_config>
                <cluster_property_set id="cib-bootstrap-options">
                    <nvpair
                        id="cib-bootstrap-options-cluster-name"
                        name="cluster-name" value="{name}"
                    />
                </cluster_property_set>
            </crm_config>
            """.format(name=name)

    @staticmethod
    def _fixture_crm_verify_success():
        return """
            <pacemaker-result api-version="2.38"
                request="crm_verify --live-check --output-as=xml"
            >
                <status code="0" message="OK"/>
            </pacemaker-result>
        """

    @staticmethod
    def _fixture_crm_verify_invalid_cib(error_list):
        errors = "\n".join(f"<error>{error}</error>" for error in error_list)
        return f"""
            <pacemaker-result api-version="2.38"
                request="crm_verify --live-check --output-as=xml"
            >
                <status code="{EXITCODE_INVALID_CIB}"
                    message="Invalid configuration"
                >
                    <errors>{errors}</errors>
                </status>
            </pacemaker-result>
        """

    def _fixture_config_crm_verify(
        self, stdout, stderr="", retval=0, cib_file=None
    ):
        self.config.runner.pcmk.verify_xml(
            cib_tempfile=cib_file,
            stdout=stdout,
            stderr=stderr,
            returncode=retval,
            env=({"CIB_file": cib_file} if cib_file else None),
        )
        self.config.fs.isfile(
            settings.pacemaker_api_result_schema,
            name="fs.exists.crm_verify_xml_schema",
        )

    def _fixture_config_live_minimal(self):
        self.config.runner.pcmk.load_state_plaintext(
            stdout="crm_mon cluster status"
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load()
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """
        )
        self._fixture_config_crm_verify(self._fixture_crm_verify_success())

    def _fixture_config_live_remote_minimal(self):
        self.config.runner.pcmk.load_state_plaintext(
            stdout="crm_mon cluster status"
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=False)
        self.config.runner.cib.load(
            optional_in_conf=self._fixture_xml_clustername("test-cib"),
            resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """,
        )
        self._fixture_config_crm_verify(self._fixture_crm_verify_success())

    def _fixture_config_local_daemons(  # noqa: PLR0913
        self,
        *,
        corosync_enabled=True,
        corosync_active=True,
        pacemaker_enabled=True,
        pacemaker_active=True,
        pacemaker_remote_enabled=False,
        pacemaker_remote_active=False,
        pcsd_enabled=True,
        pcsd_active=True,
        sbd_enabled=False,
        sbd_active=False,
    ):
        # pylint: disable=too-many-arguments
        self.config.services.is_enabled(
            "corosync",
            name="services.is_enabled.corosync",
            return_value=corosync_enabled,
        )
        self.config.services.is_running(
            "corosync",
            name="services.is_running.corosync",
            return_value=corosync_active,
        )
        self.config.services.is_enabled(
            "pacemaker",
            name="services.is_enabled.pacemaker",
            return_value=pacemaker_enabled,
        )
        self.config.services.is_running(
            "pacemaker",
            name="services.is_running.pacemaker",
            return_value=pacemaker_active,
        )
        self.config.services.is_enabled(
            "pacemaker_remote",
            name="services.is_enabled.pacemaker_remote",
            return_value=pacemaker_remote_enabled,
        )
        self.config.services.is_running(
            "pacemaker_remote",
            name="services.is_running.pacemaker_remote",
            return_value=pacemaker_remote_active,
        )
        self.config.services.is_enabled(
            "pcsd",
            name="services.is_enabled.pcsd",
            return_value=pcsd_enabled,
        )
        self.config.services.is_running(
            "pcsd",
            name="services.is_running.pcsd",
            return_value=pcsd_active,
        )
        self.config.services.is_enabled(
            "sbd",
            name="services.is_enabled.sbd_2",
            return_value=sbd_enabled,
        )
        self.config.services.is_running(
            "sbd",
            name="services.is_running.sbd_2",
            return_value=sbd_active,
        )


@mock.patch("pcs.settings.booth_enable_authfile_set_enabled", False)
@mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
@mock.patch.object(
    settings, "pacemaker_api_result_schema", rc("pcmk_rng/api/api-result.rng")
)
class FullClusterStatusPlaintext(FullClusterStatusPlaintextBase):
    # pylint: disable=too-many-public-methods
    def test_life_cib_mocked_corosync(self):
        self.config.env.set_corosync_conf_data("corosync conf data")

        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_NOT_CONSISTENT,
                    mocked_files=[file_type_codes.COROSYNC_CONF],
                    required_files=[file_type_codes.CIB],
                ),
            ],
            expected_in_processor=False,
        )

    def test_mocked_cib_life_corosync(self):
        self.config.env.set_cib_data("<cib/>")

        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_NOT_CONSISTENT,
                    mocked_files=[file_type_codes.CIB],
                    required_files=[file_type_codes.COROSYNC_CONF],
                ),
            ],
            expected_in_processor=False,
        )

    def test_fail_getting_cluster_status(self):
        self.config.runner.pcmk.load_state_plaintext(
            stdout="some stdout", stderr="some stderr", returncode=1
        )

        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    report_codes.CRM_MON_ERROR,
                    reason="some stderr\nsome stdout",
                ),
            ],
            expected_in_processor=False,
        )

    def test_fail_getting_corosync_conf(self):
        self.config.runner.pcmk.load_state_plaintext(
            stdout="crm_mon cluster status"
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load_content("invalid corosync conf")

        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.PARSE_ERROR_COROSYNC_CONF_LINE_IS_NOT_SECTION_NOR_KEY_VALUE,
                ),
            ]
        )

    def test_fail_getting_cib(self):
        self.config.runner.pcmk.load_state_plaintext(
            stdout="crm_mon cluster status",
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load()
        self.config.runner.cib.load_content(
            "some stdout", stderr="cib load error", returncode=1
        )

        self.env_assist.assert_raise_library_error(
            lambda: status.full_cluster_status_plaintext(
                self.env_assist.get_env()
            ),
            [
                fixture.error(
                    report_codes.CIB_LOAD_ERROR,
                    reason="cib load error",
                ),
            ],
            expected_in_processor=False,
        )

    def test_success_live(self):
        self._fixture_config_live_minimal()
        self._fixture_config_local_daemons()
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_success_live_verbose(self):
        self.config.env.set_known_nodes(self.node_name_list)
        self.config.runner.pcmk.can_fence_history_status(stderr="not supported")
        self.config.runner.pcmk.load_state_plaintext(
            verbose=True, stdout="crm_mon cluster status"
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load(node_name_list=self.node_name_list)
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """
        )
        self._fixture_config_crm_verify(self._fixture_crm_verify_success())
        self.config.runner.pcmk.load_ticket_state_plaintext(
            stdout="ticket status"
        )
        self._fixture_config_local_daemons()
        self.config.http.host.check_reachability(
            node_labels=self.node_name_list
        )
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(), verbose=True
            ),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  ticket status

                PCSD Status:
                  node1: Online
                  node2: Online
                  node3: Online

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_success_live_remote_node(self):
        self._fixture_config_live_remote_minimal()
        self._fixture_config_local_daemons(
            corosync_enabled=False,
            corosync_active=False,
            pacemaker_enabled=False,
            pacemaker_active=False,
            pacemaker_remote_enabled=True,
            pacemaker_remote_active=True,
        )
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test-cib
                crm_mon cluster status

                Daemon Status:
                  corosync: inactive/disabled
                  pacemaker: inactive/disabled
                  pacemaker_remote: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_success_live_remote_node_verbose(self):
        self.config.runner.pcmk.can_fence_history_status(stderr="not supported")
        self.config.runner.pcmk.load_state_plaintext(
            verbose=True, stdout="crm_mon cluster status"
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=False)
        self.config.runner.cib.load(
            optional_in_conf=self._fixture_xml_clustername("test-cib"),
            resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """,
        )
        self._fixture_config_crm_verify(self._fixture_crm_verify_success())
        self.config.runner.pcmk.load_ticket_state_plaintext(
            stdout="ticket status"
        )
        self._fixture_config_local_daemons(
            corosync_enabled=False,
            corosync_active=False,
            pacemaker_enabled=False,
            pacemaker_active=False,
            pacemaker_remote_enabled=True,
            pacemaker_remote_active=True,
        )
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(), verbose=True
            ),
            dedent(
                """\
                Cluster name: test-cib
                crm_mon cluster status

                Tickets:
                  ticket status

                Daemon Status:
                  corosync: inactive/disabled
                  pacemaker: inactive/disabled
                  pacemaker_remote: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_success_mocked(self):
        tmp_file = "/fake/tmp_file"
        env = dict(CIB_file=tmp_file)
        self.config.env.set_corosync_conf_data(rc_read("corosync.conf"))
        self.config.env.set_cib_data("<cib/>", cib_tempfile=tmp_file)
        self.config.runner.pcmk.load_state_plaintext(
            stdout="crm_mon cluster status", env=env
        )
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """,
            env=env,
        )
        self._fixture_config_crm_verify(
            self._fixture_crm_verify_success(), cib_file=tmp_file
        )
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status"""
            ),
        )

    def test_success_mocked_verbose(self):
        tmp_file = "/fake/tmp_file"
        env = dict(CIB_file=tmp_file)
        self.config.env.set_corosync_conf_data(rc_read("corosync.conf"))
        self.config.env.set_cib_data("<cib/>", cib_tempfile=tmp_file)
        self.config.runner.pcmk.can_fence_history_status(
            stderr="not supported", env=env
        )
        self.config.runner.pcmk.load_state_plaintext(
            verbose=True, stdout="crm_mon cluster status", env=env
        )
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """,
            env=env,
        )
        self._fixture_config_crm_verify(
            self._fixture_crm_verify_success(), cib_file=tmp_file
        )
        self.config.runner.pcmk.load_ticket_state_plaintext(
            stdout="ticket status", env=env
        )
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)
        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(), verbose=True
            ),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  ticket status"""
            ),
        )

    def test_success_verbose_inactive_and_fence_history(self):
        self.config.env.set_known_nodes(self.node_name_list)
        self.config.runner.pcmk.can_fence_history_status()
        self.config.runner.pcmk.load_state_plaintext(
            verbose=True,
            inactive=False,
            fence_history=True,
            stdout="crm_mon cluster status",
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load(node_name_list=self.node_name_list)
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """
        )
        self._fixture_config_crm_verify(self._fixture_crm_verify_success())
        self.config.runner.pcmk.load_ticket_state_plaintext(
            stdout="ticket status"
        )
        self._fixture_config_local_daemons()
        self.config.http.host.check_reachability(
            node_labels=self.node_name_list
        )
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(),
                verbose=True,
                hide_inactive_resources=True,
            ),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  ticket status

                PCSD Status:
                  node1: Online
                  node2: Online
                  node3: Online

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def _assert_success_with_ticket_status_failure(self, stderr="", msg=""):
        self.config.env.set_known_nodes(self.node_name_list)
        self.config.runner.pcmk.can_fence_history_status(stderr="not supported")
        self.config.runner.pcmk.load_state_plaintext(
            verbose=True, stdout="crm_mon cluster status"
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load(node_name_list=self.node_name_list)
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """
        )
        self._fixture_config_crm_verify(self._fixture_crm_verify_success())
        self.config.runner.pcmk.load_ticket_state_plaintext(
            stdout="ticket stdout", stderr=stderr, returncode=1
        )
        self._fixture_config_local_daemons()
        self.config.http.host.check_reachability(
            node_labels=self.node_name_list
        )
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(), verbose=True
            ),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  WARNING: Unable to get information about tickets{msg}

                PCSD Status:
                  node1: Online
                  node2: Online
                  node3: Online

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ).format(msg=msg),
        )

    def test_success_with_ticket_status_failure(self):
        self._assert_success_with_ticket_status_failure()

    def test_success_with_ticket_status_failure_with_message(self):
        self._assert_success_with_ticket_status_failure(
            stderr="ticket status error\nmultiline\n",
            msg="\n    ticket status error\n    multiline",
        )

    def test_crm_verify_messages(self):
        errors = [
            "error: Resource chr:0 is of type systemd and therefore cannot be used as a promotable clone resource",
            "error: Ignoring &lt;clone&gt; resource 'chr-clone' because configuration is invalid",
            "error: CIB did not pass schema validation",
            "Configuration invalid (with errors)",
        ]
        self.config.runner.pcmk.load_state_plaintext(
            stdout="crm_mon cluster status"
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load()
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """
        )
        self._fixture_config_crm_verify(
            self._fixture_crm_verify_invalid_cib(errors),
            retval=EXITCODE_INVALID_CIB,
        )
        self._fixture_config_local_daemons()
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99

                WARNINGS:
                error: Resource chr:0 is of type systemd and therefore cannot be used as a promotable clone resource
                error: Ignoring <clone> resource 'chr-clone' because configuration is invalid
                error: CIB did not pass schema validation
                Configuration invalid (with errors)

                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_crm_verify_error(self):
        self.config.runner.pcmk.load_state_plaintext(
            stdout="crm_mon cluster status"
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load()
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """
        )
        self.config.runner.pcmk.verify_xml(
            stdout="not a xml", stderr="some message"
        )
        self._fixture_config_local_daemons()
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )
        self.env_assist.assert_reports(
            [
                fixture.debug(
                    report_codes.BAD_PCMK_API_RESPONSE_FORMAT,
                    reason=(
                        "Start tag expected, '<' not found, line 1, column 1 "
                        "(<string>, line 1)"
                    ),
                    api_response="some message\nnot a xml",
                )
            ]
        )

    def test_stonith_warnings_regarding_devices_configuration(self):
        self.config.runner.pcmk.load_state_plaintext(
            stdout="crm_mon cluster status"
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load()
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="S1" class="stonith" type="fence_dummy">
                        <instance_attributes>
                            <nvpair name="method" value="cycle" />
                        </instance_attributes>
                    </primitive>
                    <primitive id="S2" class="stonith" type="fence_dummy">
                        <instance_attributes>
                            <nvpair name="action" value="value" />
                            <nvpair name="method" value="not-cycle" />
                        </instance_attributes>
                    </primitive>
                    <primitive id="S3" class="stonith" type="fence_dummy">
                        <instance_attributes>
                            <nvpair name="method" value="cycle" />
                            <nvpair name="action" value="value" />
                        </instance_attributes>
                    </primitive>
                    <primitive id="S4" class="stonith" type="fence_sbd">
                        <instance_attributes>
                            <nvpair name="method" value="cycle" />
                        </instance_attributes>
                    </primitive>
                    <primitive id="S5" class="stonith" type="fence_heuristics_ping">
                        <instance_attributes>
                            <nvpair name="method" value="cycle" />
                        </instance_attributes>
                    </primitive>
                </resources>
            """
        )
        self._fixture_config_crm_verify(self._fixture_crm_verify_success())
        self._fixture_config_local_daemons()
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99

                WARNINGS:
                Following stonith devices have the 'action' option set, it is recommended to set 'pcmk_off_action', 'pcmk_reboot_action' instead: 'S2', 'S3'
                Following stonith devices have the 'method' option set to 'cycle' which is potentially dangerous, please consider using 'onoff': 'S1', 'S3'

                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_pcsd_status_issues(self):
        self.node_name_list = ["node1", "node2", "node3", "node4", "node5"]
        self.config.env.set_known_nodes(self.node_name_list[1:])
        self.config.runner.pcmk.can_fence_history_status(stderr="not supported")
        self.config.runner.pcmk.load_state_plaintext(
            verbose=True, stdout="crm_mon cluster status"
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load(node_name_list=self.node_name_list)
        self.config.runner.cib.load(
            resources="""
                <resources>
                    <primitive id="S" class="stonith" type="fence_dummy" />
                </resources>
            """
        )
        self._fixture_config_crm_verify(self._fixture_crm_verify_success())
        self.config.runner.pcmk.load_ticket_state_plaintext(
            stdout="ticket status"
        )
        self._fixture_config_local_daemons()
        self.config.http.host.check_reachability(
            communication_list=[
                # node1 has no record in known-hosts
                dict(
                    label="node2",
                    was_connected=False,
                    errno=7,
                    error_msg="node2 error",
                ),
                dict(label="node3", response_code=401, output="node3 output"),
                dict(label="node4", response_code=500, output="node4 output"),
                dict(label="node5"),
            ]
        )
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(
                self.env_assist.get_env(), verbose=True
            ),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Tickets:
                  ticket status

                PCSD Status:
                  node1: Unable to authenticate
                  node2: Offline
                  node3: Unable to authenticate
                  node4: Online
                  node5: Online

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ),
        )

    def test_daemon_status_all_on(self):
        self._fixture_config_live_minimal()
        self._fixture_config_local_daemons(
            corosync_enabled=True,
            corosync_active=True,
            pacemaker_enabled=True,
            pacemaker_active=True,
            pacemaker_remote_enabled=True,
            pacemaker_remote_active=True,
            pcsd_enabled=True,
            pcsd_active=True,
            sbd_enabled=True,
            sbd_active=True,
        )
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)
        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pacemaker_remote: active/enabled
                  pcsd: active/enabled
                  sbd: active/enabled"""
            ),
        )

    def test_daemon_status_all_off(self):
        self._fixture_config_live_minimal()
        self._fixture_config_local_daemons(
            corosync_enabled=False,
            corosync_active=False,
            pacemaker_enabled=False,
            pacemaker_active=False,
            pacemaker_remote_enabled=False,
            pacemaker_remote_active=False,
            pcsd_enabled=False,
            pcsd_active=False,
            sbd_enabled=False,
            sbd_active=False,
        )
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)
        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99
                crm_mon cluster status

                Daemon Status:
                  corosync: inactive/disabled
                  pacemaker: inactive/disabled
                  pcsd: inactive/disabled"""
            ),
        )

    def test_move_constrains_warnings(self):
        self.config.runner.pcmk.load_state_plaintext(
            stdout="crm_mon cluster status",
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load()
        self.config.runner.cib.load(
            constraints="""
            <constraints>
                <rsc_location id="cli-ban-P1-on-node1" rsc="P1"
                    role="Started" node="node1" score="-INFINITY"/>
                <rsc_location id="cli-prefer-P1" rsc="P1" role="Started"
                    node="node3" score="INFINITY"/>
                <rsc_location id="cli-prefer-P2" rsc="P2" role="Started"
                    node="node1" score="INFINITY"/>
                <rsc_location id="cli-ban-P2-on-node1" rsc="P2"
                    role="Started" node="node1" score="-INFINITY"/>
                <rsc_location id="location-P3-node3--INFINITY" rsc="P3"
                    node="node3" score="-INFINITY"/>
            </constraints>
            """,
            resources="""
            <resources>
                <primitive class="ocf" id="P1" provider="pacemaker"
                    type="Dummy"/>
                <primitive class="ocf" id="P2" provider="pacemaker"
                    type="Dummy"/>
                <primitive class="ocf" id="P3" provider="pacemaker"
                    type="Dummy"/>
            </resources>
            """,
        )
        self._fixture_config_crm_verify(self._fixture_crm_verify_success())
        self._fixture_config_local_daemons(sbd_enabled=True, sbd_active=True)
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99

                WARNINGS:
                Following resources have been moved and their move constraints are still in place: 'P1', 'P2'
                Run 'pcs constraint location' or 'pcs resource clear <resource id>' to view or remove the constraints, respectively

                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled
                  sbd: active/enabled"""
            ),
        )

    def test_expired_move_constraints_warnings(self):
        self.config.runner.pcmk.load_state_plaintext(
            stdout="crm_mon cluster status",
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load()
        self.config.runner.cib.load(
            constraints="""
            <constraints>
                <rsc_location id="cli-ban-P1-on-node1" rsc="P1"
                    role="Started" node="node1" score="-INFINITY"/>
                <rsc_location id="cli-prefer-P2" rsc="P2" role="Started"
                    node="node1" score="INFINITY">
                    <rule id="cli-prefer-rule-P2" score="INFINITY" boolean-op="and">
                        <expression id="cli-prefer-expr-P2" attribute="#uname"
                                    operation="eq" value="P2" type="string"/>
                        <date_expression id="cli-prefer-lifetime-end-P2"
                                    operation="lt" end="0000-01-1 01:00:00 +02:00"/>
                    </rule>
                </rsc_location>
            </constraints>
            """,
            resources="""
            <resources>
                <primitive class="ocf" id="P1" provider="pacemaker"
                    type="Dummy"/>
                <primitive class="ocf" id="P2" provider="pacemaker"
                    type="Dummy"/>
                <primitive class="ocf" id="P3" provider="pacemaker"
                    type="Dummy"/>
            </resources>
            """,
        )
        self._fixture_config_crm_verify(self._fixture_crm_verify_success())
        self._fixture_config_local_daemons(sbd_enabled=True, sbd_active=True)
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)
        self.config.runner.pcmk.get_rule_in_effect_status(
            "cli-prefer-rule-P2", returncode=RULE_EXPIRED_RETURNCODE
        )

        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99

                WARNINGS:
                Following resources have been moved and their move constraints are still in place: 'P1'
                Run 'pcs constraint location' or 'pcs resource clear <resource id>' to view or remove the constraints, respectively

                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled
                  sbd: active/enabled"""
            ),
        )

    def test_expired_and_in_effect_move_constraints_warnings(self):
        self.config.runner.pcmk.load_state_plaintext(
            stdout="crm_mon cluster status",
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load()
        self.config.runner.cib.load(
            constraints="""
            <constraints>
                <rsc_location id="cli-prefer-P1" rsc="P1" role="Started"
                    node="node1" score="INFINITY">
                    <rule id="cli-prefer-rule-P1" score="INFINITY" boolean-op="and">
                        <expression id="cli-prefer-expr-P1" attribute="#uname"
                                    operation="eq" value="P1" type="string"/>
                        <date_expression id="cli-prefer-lifetime-end-P1"
                                    operation="lt" end="0000-01-1 01:00:00 +02:00"/>
                    </rule>
                </rsc_location>
                <rsc_location id="cli-prefer-P2" rsc="P2" role="Started"
                    node="node2" score="INFINITY">
                    <rule id="cli-prefer-rule-P2" score="INFINITY" boolean-op="and">
                        <expression id="cli-prefer-expr-P2" attribute="#uname"
                                    operation="eq" value="P2" type="string"/>
                        <date_expression id="cli-prefer-lifetime-end-P2"
                                    operation="lt" end="0000-01-1 01:00:00 +02:00"/>
                    </rule>
                </rsc_location>
            </constraints>
            """,
            resources="""
            <resources>
                <primitive class="ocf" id="P1" provider="pacemaker"
                    type="Dummy"/>
                <primitive class="ocf" id="P2" provider="pacemaker"
                    type="Dummy"/>
                <primitive class="ocf" id="P3" provider="pacemaker"
                    type="Dummy"/>
            </resources>
            """,
        )
        self._fixture_config_crm_verify(self._fixture_crm_verify_success())
        self._fixture_config_local_daemons(sbd_enabled=True, sbd_active=True)
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)
        self.config.runner.pcmk.get_rule_in_effect_status(
            "cli-prefer-rule-P1",
            returncode=RULE_EXPIRED_RETURNCODE,
            name="runner.pcmk.get_rule_in_effect_status-1",
        )
        self.config.runner.pcmk.get_rule_in_effect_status(
            "cli-prefer-rule-P2",
            returncode=RULE_IN_EFFECT_RETURNCODE,
            name="runner.pcmk.get_rule_in_effect_status-2",
        )

        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99

                WARNINGS:
                Following resources have been moved and their move constraints are still in place: 'P2'
                Run 'pcs constraint location' or 'pcs resource clear <resource id>' to view or remove the constraints, respectively

                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled
                  sbd: active/enabled"""
            ),
        )

    def test_bundle_warnings(self):
        self.config.runner.pcmk.load_state_plaintext(
            stdout="crm_mon cluster status",
        )
        self.config.fs.exists(settings.corosync_conf_file, return_value=True)
        self.config.corosync_conf.load()
        self.config.runner.cib.load(
            resources="""
            <resources>
                <bundle id="bundle-bad">
                    <rkt image="pcs:test" />
                </bundle>
                <bundle id="bundle-good">
                    <docker image="pcs:test" />
                </bundle>
            </resources>
            """,
        )
        self._fixture_config_crm_verify(self._fixture_crm_verify_success())
        self._fixture_config_local_daemons(sbd_enabled=True, sbd_active=True)
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)

        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99

                WARNINGS:
                Bundle 'bundle-bad' uses unsupported container type. Supported container types are: 'docker', 'podman'

                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled
                  sbd: active/enabled"""
            ),
        )


class FullClusterStatusPlaintextBoothWarning(FullClusterStatusPlaintextBase):
    def setUp(self):
        super().setUp()
        self.settings_patcher = mock.patch(
            "pcs.settings.pacemaker_api_result_schema",
            rc("pcmk_rng/api/api-result.rng"),
        )
        self.settings_patcher.start()
        self._fixture_config_live_minimal()
        self._fixture_config_local_daemons()

    def tearDown(self):
        self.settings_patcher.stop()

    def _assert_status_output(self, warning=None):
        warning_str = ""
        if warning:
            warning_str = f"\n\nWARNINGS:\n{warning}\n"
        self.assertEqual(
            status.full_cluster_status_plaintext(self.env_assist.get_env()),
            dedent(
                """\
                Cluster name: test99{warning_str}
                crm_mon cluster status

                Daemon Status:
                  corosync: active/enabled
                  pacemaker: active/enabled
                  pcsd: active/enabled"""
            ).format(warning_str=warning_str),
        )

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
    def test_booth_not_configured_set_enabled(self):
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            _booth_config_path_fixture(),
            exists=False,
        )
        self._assert_status_output()

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
    def test_booth_authfile_not_configured_set_enabled(self):
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            _booth_config_path_fixture(),
            exists=True,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            _booth_config_path_fixture(),
            content=dedent(
                """
                site = 1.1.1.1
                """
            ).encode("utf-8"),
        )
        self._assert_status_output()

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", False)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", True)
    def test_booth_not_configured_unset_enabled(self):
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            _booth_config_path_fixture(),
            exists=False,
        )
        self._assert_status_output()

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
    def test_missing_option(self):
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            _booth_config_path_fixture(),
            exists=True,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            _booth_config_path_fixture(),
            content=dedent(
                f"""
                authfile = {_booth_config_path_fixture()}
                site = 1.1.1.1
                """
            ).encode("utf-8"),
        )
        self._assert_status_output(
            "Booth is configured to use an authfile, but authfile is not "
            "enabled. Run 'pcs booth enable-authfile --name booth' to enable "
            "usage of booth autfile."
        )

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
    def test_properly_configured(self):
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            _booth_config_path_fixture(),
            exists=True,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            _booth_config_path_fixture(),
            content=dedent(
                f"""
                authfile = {_booth_config_path_fixture()}
                {constants.AUTHFILE_FIX_OPTION} = yes
                site = 1.1.1.1
                """
            ).encode("utf-8"),
        )
        self._assert_status_output()

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", False)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", True)
    def test_unsupported_option(self):
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            _booth_config_path_fixture(),
            exists=True,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            _booth_config_path_fixture(),
            content=dedent(
                f"""
                authfile = {_booth_config_path_fixture()}
                {constants.AUTHFILE_FIX_OPTION} = yes
                site = 1.1.1.1
                """
            ).encode("utf-8"),
        )
        self._assert_status_output(
            "Unsupported option 'enable-authfile' is set in booth "
            "configuration. Run 'pcs booth enable-booth-clean --name booth' to "
            "remove the option."
        )

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", False)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", True)
    def test_unsupported_option_not_present(self):
        self.config.fs.isfile(settings.crm_rule_exec, return_value=True)
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            _booth_config_path_fixture(),
            exists=True,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            _booth_config_path_fixture(),
            content=dedent(
                f"""
                authfile = {_booth_config_path_fixture()}
                site = 1.1.1.1
                """
            ).encode("utf-8"),
        )
        self._assert_status_output()


def _fixture_primitive_resource_dto(
    resource_id: str,
    resource_agent: str,
    target_role: Optional[PcmkRoleType] = None,
    managed: bool = True,
) -> PrimitiveStatusDto:
    return PrimitiveStatusDto(
        resource_id=resource_id,
        instance_id=None,
        resource_agent=resource_agent,
        role=PCMK_STATUS_ROLE_STOPPED,
        target_role=target_role,
        active=False,
        orphaned=False,
        blocked=False,
        maintenance=False,
        description=None,
        failed=False,
        managed=managed,
        failure_ignored=False,
        node_names=[],
        pending=None,
        locked_to=None,
    )


@mock.patch.object(
    settings,
    "pacemaker_api_result_schema",
    rc("pcmk_rng/api/api-result.rng"),
)
class ResourcesStatus(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_empty_resources(self):
        self.config.runner.pcmk.load_state()

        result = status.resources_status(self.env_assist.get_env())
        self.assertEqual(result, ResourcesStatusDto([]))

    def test_bad_xml_format(self):
        self.config.runner.pcmk.load_state(
            resources="""
                <resources>
                    <resource />
                </resources>
            """,
        )

        self.env_assist.assert_raise_library_error(
            lambda: status.resources_status(
                self.env_assist.get_env(),
            ),
            [fixture.error(report_codes.BAD_CLUSTER_STATE_FORMAT)],
            False,
        )

    def test_bad_xml(self):
        self.config.runner.pcmk.load_state(
            resources="""
                <resources>
                    <resource id="R7" resource_agent="ocf:pacemaker:Dummy"
                        role="NotPcmkRole" active="false" orphaned="false"
                        blocked="false" maintenance="false" managed="true"
                        failed="false" failure_ignored="false"
                        nodes_running_on="0"
                    />
                </resources>
            """,
        )

        self.env_assist.assert_raise_library_error(
            lambda: status.resources_status(self.env_assist.get_env()),
            [
                fixture.error(
                    report_codes.BAD_CLUSTER_STATE_DATA,
                    reason="Resource 'R7' contains an unknown role 'NotPcmkRole'",
                )
            ],
            False,
        )

    def test_all_resources(self):
        self.config.runner.pcmk.load_state(
            filename=rc("crm_mon.all_resources.xml")
        )

        result = status.resources_status(self.env_assist.get_env())

        self.assertEqual(
            result,
            ResourcesStatusDto(
                [
                    BundleStatusDto(
                        resource_id="B1",
                        type="docker",
                        image="pcs:test",
                        unique=True,
                        maintenance=False,
                        description=None,
                        managed=False,
                        failed=False,
                        replicas=[
                            BundleReplicaStatusDto(
                                replica_id="0",
                                member=None,
                                remote=None,
                                container=_fixture_primitive_resource_dto(
                                    "B1-docker-0",
                                    "ocf:heartbeat:docker",
                                    target_role=PCMK_ROLE_STOPPED,
                                    managed=False,
                                ),
                                ip_address=_fixture_primitive_resource_dto(
                                    "B1-ip-192.168.100.200",
                                    "ocf:heartbeat:IPaddr2",
                                    target_role=PCMK_ROLE_STOPPED,
                                    managed=False,
                                ),
                            ),
                            BundleReplicaStatusDto(
                                replica_id="1",
                                member=None,
                                remote=None,
                                container=_fixture_primitive_resource_dto(
                                    "B1-docker-1",
                                    "ocf:heartbeat:docker",
                                    target_role=PCMK_ROLE_STOPPED,
                                    managed=False,
                                ),
                                ip_address=_fixture_primitive_resource_dto(
                                    "B1-ip-192.168.100.201",
                                    "ocf:heartbeat:IPaddr2",
                                    target_role=PCMK_ROLE_STOPPED,
                                    managed=False,
                                ),
                            ),
                            BundleReplicaStatusDto(
                                replica_id="2",
                                member=None,
                                remote=None,
                                container=_fixture_primitive_resource_dto(
                                    "B1-docker-2",
                                    "ocf:heartbeat:docker",
                                    target_role=PCMK_ROLE_STOPPED,
                                    managed=False,
                                ),
                                ip_address=_fixture_primitive_resource_dto(
                                    "B1-ip-192.168.100.202",
                                    "ocf:heartbeat:IPaddr2",
                                    target_role=PCMK_ROLE_STOPPED,
                                    managed=False,
                                ),
                            ),
                            BundleReplicaStatusDto(
                                replica_id="3",
                                member=None,
                                remote=None,
                                container=_fixture_primitive_resource_dto(
                                    "B1-docker-3",
                                    "ocf:heartbeat:docker",
                                    target_role=PCMK_ROLE_STOPPED,
                                    managed=False,
                                ),
                                ip_address=_fixture_primitive_resource_dto(
                                    "B1-ip-192.168.100.203",
                                    "ocf:heartbeat:IPaddr2",
                                    target_role=PCMK_ROLE_STOPPED,
                                    managed=False,
                                ),
                            ),
                        ],
                    ),
                    _fixture_primitive_resource_dto(
                        "R7", "ocf:pacemaker:Dummy"
                    ),
                    _fixture_primitive_resource_dto(
                        "S2", "stonith:fence_kdump"
                    ),
                    GroupStatusDto(
                        resource_id="G2",
                        instance_id=None,
                        maintenance=False,
                        description=None,
                        managed=True,
                        disabled=False,
                        members=[
                            _fixture_primitive_resource_dto(
                                "R5", "ocf:pacemaker:Dummy"
                            ),
                            _fixture_primitive_resource_dto(
                                "S1", "stonith:fence_kdump"
                            ),
                        ],
                    ),
                    CloneStatusDto(
                        resource_id="G1-clone",
                        multi_state=True,
                        unique=False,
                        maintenance=False,
                        description=None,
                        managed=True,
                        disabled=False,
                        failed=False,
                        failure_ignored=False,
                        target_role=None,
                        instances=[
                            GroupStatusDto(
                                resource_id="G1",
                                instance_id="0",
                                maintenance=False,
                                description=None,
                                managed=True,
                                disabled=False,
                                members=[
                                    _fixture_primitive_resource_dto(
                                        "R2", "ocf:pacemaker:Stateful"
                                    ),
                                    _fixture_primitive_resource_dto(
                                        "R3", "ocf:pacemaker:Stateful"
                                    ),
                                    _fixture_primitive_resource_dto(
                                        "R4", "ocf:pacemaker:Stateful"
                                    ),
                                ],
                            )
                        ],
                    ),
                    CloneStatusDto(
                        resource_id="R6-clone",
                        multi_state=False,
                        unique=False,
                        maintenance=False,
                        description=None,
                        managed=True,
                        disabled=False,
                        failed=False,
                        failure_ignored=False,
                        target_role=None,
                        instances=[
                            _fixture_primitive_resource_dto(
                                "R6", "ocf:pacemaker:Dummy"
                            )
                        ],
                    ),
                ]
            ),
        )

    def test_bundle_skip(self):
        self.config.runner.pcmk.load_state(
            resources="""
                <resources>
                    <bundle id="B1" type="docker" image="pcs:test" unique="true"
                        maintenance="false" managed="false" failed="false"
                    >
                        <replica id="0">
                            <resource id="B1-0"
                                resource_agent="ocf:heartbeat:Dummy"
                                role="Stopped" target_role="Stopped"
                                active="false" orphaned="false" blocked="false"
                                maintenance="false" managed="true" failed="false"
                                failure_ignored="false" nodes_running_on="0"
                            />
                            <resource id="B1-docker-0"
                                resource_agent="ocf:heartbeat:docker"
                                role="Stopped" target_role="Stopped"
                                active="false" orphaned="false" blocked="false"
                                maintenance="false" managed="false" failed="false"
                                failure_ignored="false" nodes_running_on="0"
                            />
                            <resource id="B1-0"
                                resource_agent="ocf:pacemaker:remote"
                                role="Stopped" target_role="Stopped"
                                active="false" orphaned="false" blocked="false"
                                maintenance="false" managed="false" failed="false"
                                failure_ignored="false" nodes_running_on="0"
                            />
                        </replica>
                    </bundle>
                </resources>
            """,
        )

        result = status.resources_status(self.env_assist.get_env())
        self.assertEqual(result, ResourcesStatusDto([]))
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.CLUSTER_STATUS_BUNDLE_MEMBER_ID_AS_IMPLICIT,
                    bundle_id="B1",
                    bad_ids=["B1-0"],
                )
            ]
        )



from typing import Mapping, cast

from pcs.common import reports
from pcs.common.auth import HostAuthData, HostWithTokenAuthData
from pcs.common.file import RawFileError
from pcs.common.node_communicator import PcsKnownHost, RequestTarget
from pcs.common.types import StringSequence
from pcs.lib.auth import validations
from pcs.lib.communication.nodes import Auth
from pcs.lib.communication.tools import run
from pcs.lib.env import LibraryEnvironment, LibraryError
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.raw_file import raw_file_error_report
from pcs.lib.host.config.facade import Facade as KnownHostsFacade
from pcs.lib.interface.config import ParserErrorException
from pcs.lib.node import get_existing_nodes_names
from pcs.lib.pcs_cfgsync.save_sync import save_sync_new_known_hosts


def auth_hosts_token_no_sync(
    env: LibraryEnvironment, hosts: Mapping[str, HostWithTokenAuthData]
) -> None:
    """
    Add the specified hosts with custom tokens to the local known-hosts file.
    The updates to the known-hosts file are not synchronized into the cluster.
    Automatic known hosts synchronization should be disabled in cluster when
    setting a custom token.

    hosts -- mapping of host names to data needed for their authentication
    """
    if env.report_processor.report_list(
        validations.validate_hosts_with_token(hosts)
    ).has_errors:
        raise LibraryError()

    new_known_hosts = [
        host_data.to_known_host(host_name)
        for host_name, host_data in hosts.items()
    ]

    known_hosts_facade = _read_known_hosts_file(env.report_processor)

    known_hosts_facade.update_known_hosts(new_known_hosts)
    known_hosts_facade.set_data_version(known_hosts_facade.data_version + 1)

    try:
        FileInstance.for_known_hosts().write_facade(
            known_hosts_facade, can_overwrite=True
        )
    except RawFileError as e:
        env.report_processor.report(raw_file_error_report(e))
        raise LibraryError() from e


def auth_hosts(  # noqa: PLR0912
    env: LibraryEnvironment, hosts: Mapping[str, HostAuthData]
) -> None:
    """
    Authenticate the specified hosts. Sync the known-hosts file to all cluster
    nodes if the local node is in a cluster

    hosts -- mapping of host names to data needed for their authentication
    """
    if env.report_processor.report_list(
        validations.validate_hosts(hosts)
    ).has_errors:
        raise LibraryError()

    node_communicator = env.get_node_communicator()
    com_cmd = Auth(hosts, env.report_processor)
    # we do not want to raise LibraryError in case only some nodes returned
    # errors, since we want to update the known-hosts file with whatever tokens
    # we were able to receive - this is how the old impl behaved
    # this means we cannot blindly check errors by using processor.has_errors
    received_tokens: dict[str, str] = run(
        node_communicator,
        com_cmd,
    )  # type: ignore[no-untyped-call]

    new_known_hosts = [
        PcsKnownHost(
            name=host_name,
            token=received_tokens[host_name],
            dest_list=auth_data.dest_list,
        )
        for host_name, auth_data in hosts.items()
        if host_name in received_tokens
    ]

    if not new_known_hosts:
        if env.report_processor.has_errors:
            raise LibraryError()
        return

    known_hosts_facade = _read_known_hosts_file(env.report_processor)

    if not FileInstance.for_corosync_conf().raw_file.exists():
        # we are not running in a cluster, so just save the new tokens locally
        known_hosts_facade.update_known_hosts(new_known_hosts)
        known_hosts_facade.set_data_version(known_hosts_facade.data_version + 1)

        try:
            FileInstance.for_known_hosts().write_facade(
                known_hosts_facade, can_overwrite=True
            )
        except RawFileError as e:
            env.report_processor.report(raw_file_error_report(e))
            raise LibraryError() from e
        return

    # we are in cluster, so we distribute the new tokens
    corosync_conf = env.get_corosync_conf()
    cluster_name = corosync_conf.get_cluster_name()

    # we want to send the tokens to all cluster nodes, but we want to use
    # the new tokens in case we ran auth on any nodes that are already in
    # the cluster
    node_names, report_list = get_existing_nodes_names(corosync_conf, None)
    env.report_processor.report_list(report_list)
    new_hosts_already_in_cluster = set(received_tokens) & set(node_names)
    (
        report_list,
        target_list,
    ) = env.get_node_target_factory().get_target_list_with_reports(
        sorted(set(node_names) - new_hosts_already_in_cluster),
        allow_skip=False,
        report_none_host_found=False,
    )
    env.report_processor.report_list(report_list)
    target_list.extend(
        RequestTarget.from_known_host(host)
        for host in new_known_hosts
        if host.name in new_hosts_already_in_cluster
    )
    if not target_list:
        # we can end, since we have no cluster nodes where to send the config to
        if env.report_processor.has_errors:
            raise LibraryError()
        return

    conflict_detected, new_file = save_sync_new_known_hosts(
        known_hosts_facade,
        new_known_hosts,
        [],
        cluster_name,
        target_list,
        node_communicator,
        env.report_processor,
    )
    try:
        if new_file is not None:
            FileInstance.for_known_hosts().write_facade(new_file, True)
    except RawFileError as e:
        env.report_processor.report(raw_file_error_report(e))

    if conflict_detected:
        env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.PcsCfgsyncConflictRepeatAction()
            )
        )
    if env.report_processor.has_errors:
        raise LibraryError()


def deauth_hosts(env: LibraryEnvironment, hosts: StringSequence) -> None:
    """
    Deauth the specified hosts. Sync the known-hosts file to all cluster
    nodes if the local node is in a cluster

    hosts -- host names of hosts to be deauthenticated
    """
    if not hosts:
        env.report_processor.report(
            reports.ReportItem.error(reports.messages.NoHostSpecified())
        )
        raise LibraryError()

    known_hosts_facade = _read_known_hosts_file(env.report_processor)

    current_known_hosts = known_hosts_facade.known_hosts
    unknown_hosts = {
        host_name for host_name in hosts if host_name not in current_known_hosts
    }
    if unknown_hosts:
        env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.HostNotFound(sorted(unknown_hosts))
            )
        )
    if not (set(hosts) - unknown_hosts):
        if env.report_processor.has_errors:
            raise LibraryError()
        return

    _deauth_hosts_common(env, known_hosts_facade, hosts)


def deauth_all_local_hosts(env: LibraryEnvironment) -> None:
    """
    Deauth all hosts present in the local known-hosts file. Sync the known-hosts
    file to all cluster nodes if the local node is in a cluster
    """
    known_hosts_facade = _read_known_hosts_file(env.report_processor)

    all_known_hosts = list(known_hosts_facade.known_hosts.keys())
    if not all_known_hosts:
        return

    _deauth_hosts_common(env, known_hosts_facade, all_known_hosts)


def _read_known_hosts_file(
    report_processor: reports.ReportProcessor,
) -> KnownHostsFacade:
    file_instance = FileInstance.for_known_hosts()
    try:
        if file_instance.raw_file.exists():
            return cast(KnownHostsFacade, file_instance.read_to_facade())
        return KnownHostsFacade.create(data_version=0)
    except RawFileError as e:
        report_processor.report(raw_file_error_report(e))
    except ParserErrorException as e:
        report_processor.report_list(
            file_instance.parser_exception_to_report_list(e)
        )
    raise LibraryError()


def _deauth_hosts_common(
    env: LibraryEnvironment,
    known_hosts_facade: KnownHostsFacade,
    hosts_to_deauth: StringSequence,
) -> None:
    """
    Extracted common parts of deauth commands, to reduce code duplication
    Deauth hosts and sync the known-hosts file if needed.
    """
    if not FileInstance.for_corosync_conf().raw_file.exists():
        known_hosts_facade.remove_known_hosts(hosts_to_deauth)
        known_hosts_facade.set_data_version(known_hosts_facade.data_version + 1)
        try:
            FileInstance.for_known_hosts().write_facade(
                known_hosts_facade, can_overwrite=True
            )
        except RawFileError as e:
            env.report_processor.report(raw_file_error_report(e))
        if env.report_processor.has_errors:
            raise LibraryError()
        return

    node_communicator = env.get_node_communicator()
    corosync_conf = env.get_corosync_conf()
    cluster_name = corosync_conf.get_cluster_name()
    node_names, report_list = get_existing_nodes_names(corosync_conf, None)
    env.report_processor.report_list(report_list)
    report_list, target_list = (
        env.get_node_target_factory().get_target_list_with_reports(
            node_names, allow_skip=False
        )
    )
    env.report_processor.report_list(report_list)
    if not target_list:
        # we can end, since we have no cluster nodes where to send the config to
        if env.report_processor.has_errors:
            raise LibraryError()
        return

    conflict_detected, new_file = save_sync_new_known_hosts(
        known_hosts_facade,
        [],
        hosts_to_deauth,
        cluster_name,
        target_list,
        node_communicator,
        env.report_processor,
    )
    try:
        if new_file is not None:
            FileInstance.for_known_hosts().write_facade(new_file, True)
    except RawFileError as e:
        env.report_processor.report(raw_file_error_report(e))
    if conflict_detected:
        env.report_processor.report(
            reports.ReportItem.error(
                reports.messages.PcsCfgsyncConflictRepeatAction()
            )
        )

    if env.report_processor.has_errors:
        raise LibraryError()



from unittest import TestCase

from pcs.common import (
    const,
    reports,
)
from pcs.common.pacemaker.types import CibResourceDiscovery
from pcs.lib.cib import const as cib_const
from pcs.lib.commands.constraint import location

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


class CreatePlainWithRule(TestCase):
    # pylint: disable=too-many-public-methods
    resources_xml = """
        <resources>
          <primitive id="R1" class="ocf" provider="pacemaker" type="Dummy">
            <operations>
              <op id="R1-monitor" name="monitor" interval="60s" />
            </operations>
          </primitive>
          <group id="G1">
            <primitive id="G1R1" class="ocf" provider="pacemaker" type="Dummy" />
            <primitive id="G1R2" class="ocf" provider="pacemaker" type="Dummy" />
          </group>
          <clone id="C1">
            <primitive id="C1R1" class="ocf" provider="pacemaker" type="Dummy" />
          </clone>
          <bundle id="B1">
            <docker image="pcs:test"/>
            <primitive id="B1R1" class="ocf" provider="pacemaker" type="Dummy" />
          </bundle>
        </resources>
    """
    constraints_xml_minimal = """
        <constraints>
          <rsc_location id="location-{resource_id}" rsc="{resource_id}">
            <rule id="location-{resource_id}-rule" boolean-op="and"
                score="INFINITY"
            >
              <expression id="location-{resource_id}-rule-expr"
                  attribute="#uname" operation="eq" value="node1"
              />
            </rule>
          </rsc_location>
          {other_constraint}
        </constraints>
    """

    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_old_cib(self):
        resource_id = "R1"
        self.config.runner.cib.load(resources=self.resources_xml)
        self.config.runner.cib.upgrade()
        self.config.runner.cib.load(
            resources=self.resources_xml,
            filename="cib-empty-3.9.xml",
            name="runner.cib.load.2",
        )
        constraints_xml = self.constraints_xml_minimal.format(
            resource_id=resource_id, other_constraint=""
        )
        self.config.env.push_cib(
            constraints=constraints_xml,
            load_key="runner.cib.load.2",
        )

        location.create_plain_with_rule(
            self.env_assist.get_env(),
            const.RESOURCE_ID_TYPE_PLAIN,
            resource_id,
            "#uname eq node1",
            {},
            {},
            [],
        )
        self.env_assist.assert_reports(
            [
                fixture.info(reports.codes.CIB_UPGRADE_SUCCESSFUL),
            ]
        )

    def test_old_cib_update_not_possible(self):
        resource_id = "R1"
        self.config.runner.cib.load(resources=self.resources_xml)
        self.config.runner.cib.upgrade()
        self.config.runner.cib.load(
            resources=self.resources_xml, name="runner.cib.load.2"
        )

        self.env_assist.assert_raise_library_error(
            lambda: location.create_plain_with_rule(
                self.env_assist.get_env(),
                const.RESOURCE_ID_TYPE_PLAIN,
                resource_id,
                "#uname eq node1",
                {},
                {},
                [],
            ),
            [
                fixture.error(
                    reports.codes.CIB_UPGRADE_FAILED_TO_MINIMAL_REQUIRED_VERSION,
                    current_version="3.1",
                    required_version="3.9.0",
                )
            ],
            expected_in_processor=False,
        )

    def assert_success_minimal(self, resource_id):
        self.config.runner.cib.load(
            resources=self.resources_xml, filename="cib-empty-3.9.xml"
        )
        constraints_xml = self.constraints_xml_minimal.format(
            resource_id=resource_id, other_constraint=""
        )
        self.config.env.push_cib(constraints=constraints_xml)

        location.create_plain_with_rule(
            self.env_assist.get_env(),
            const.RESOURCE_ID_TYPE_PLAIN,
            resource_id,
            "#uname eq node1",
            {},
            {},
            [],
        )

    def test_success_minimal_primitive(self):
        self.assert_success_minimal("R1")

    def test_success_minimal_group(self):
        self.assert_success_minimal("G1")

    def test_success_minimal_clone(self):
        self.assert_success_minimal("C1")

    def test_success_minimal_bundle(self):
        self.assert_success_minimal("B1")

    def test_success_all_options_1(self):
        constraints_xml = """
            <constraints>
              <rsc_location id="L1" rsc="R1" resource-discovery="always">
                <rule id="X" boolean-op="and" score="9" role="Promoted">
                  <expression id="X-expr"
                      attribute="#uname" operation="eq" value="node1"
                  />
                </rule>
              </rsc_location>
            </constraints>
        """
        self.config.runner.cib.load(
            resources=self.resources_xml, filename="cib-empty-3.9.xml"
        )
        self.config.env.push_cib(constraints=constraints_xml)

        location.create_plain_with_rule(
            self.env_assist.get_env(),
            const.RESOURCE_ID_TYPE_PLAIN,
            "R1",
            "#uname eq node1",
            {
                "id": "X",
                # test role normalization
                "role": str(const.PCMK_ROLE_PROMOTED).upper(),
                "score": "9",
            },
            {"id": "L1", "resource-discovery": CibResourceDiscovery.ALWAYS},
            [],
        )

    def test_success_all_options_2(self):
        constraints_xml = """
            <constraints>
              <rsc_location id="L1" rsc-pattern="R\\d+"
                  resource-discovery="exclusive"
              >
                <rule id="L1-rule" boolean-op="and" score-attribute="attr"
                    role="Unpromoted"
                >
                  <expression id="L1-rule-expr"
                      attribute="#uname" operation="eq" value="node1"
                  />
                </rule>
              </rsc_location>
            </constraints>
        """
        self.config.runner.cib.load(
            resources=self.resources_xml, filename="cib-empty-3.9.xml"
        )
        self.config.env.push_cib(constraints=constraints_xml)

        location.create_plain_with_rule(
            self.env_assist.get_env(),
            const.RESOURCE_ID_TYPE_REGEXP,
            "R\\d+",
            "#uname eq node1",
            {"role": const.PCMK_ROLE_UNPROMOTED, "score-attribute": "attr"},
            {"id": "L1", "resource-discovery": CibResourceDiscovery.EXCLUSIVE},
            [],
        )

    def test_rule_with_integer(self):
        constraints_xml = """
            <constraints>
              <rsc_location id="location-R1" rsc="R1">
                <rule id="location-R1-rule" boolean-op="and" score="INFINITY">
                  <expression id="location-R1-rule-expr"
                      attribute="attr" operation="gt" type="integer" value="7"
                  />
                </rule>
              </rsc_location>
            </constraints>
        """
        self.config.runner.cib.load(
            resources=self.resources_xml, filename="cib-empty-3.9.xml"
        )
        self.config.env.push_cib(constraints=constraints_xml)

        location.create_plain_with_rule(
            self.env_assist.get_env(),
            const.RESOURCE_ID_TYPE_PLAIN,
            "R1",
            "attr gt integer 7",
            {},
            {},
            [],
        )

    def test_duplicate_constraint(self):
        self.config.runner.cib.load(
            resources=self.resources_xml,
            constraints=self.constraints_xml_minimal.format(
                resource_id="R1", other_constraint=""
            ),
            filename="cib-empty-3.9.xml",
        )
        self.env_assist.assert_raise_library_error(
            lambda: location.create_plain_with_rule(
                self.env_assist.get_env(),
                const.RESOURCE_ID_TYPE_PLAIN,
                "R1",
                "#uname eq node1",
                {},
                {},
                [],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.DUPLICATE_CONSTRAINTS_EXIST,
                    force_code=reports.codes.FORCE,
                    constraint_ids=["location-R1"],
                )
            ]
        )

    def test_duplicate_constraint_forced(self):
        self.config.runner.cib.load(
            resources=self.resources_xml,
            constraints=self.constraints_xml_minimal.format(
                resource_id="R1", other_constraint=""
            ),
            filename="cib-empty-3.9.xml",
        )
        constraints_xml = self.constraints_xml_minimal.format(
            resource_id="R1",
            other_constraint="""
                <rsc_location id="location-R1-1" rsc="R1">
                  <rule id="location-R1-1-rule" boolean-op="and" score="INFINITY">
                    <expression id="location-R1-1-rule-expr"
                        attribute="#uname" operation="eq" value="node1"
                    />
                  </rule>
                </rsc_location>
            """,
        )
        self.config.env.push_cib(constraints=constraints_xml)

        location.create_plain_with_rule(
            self.env_assist.get_env(),
            const.RESOURCE_ID_TYPE_PLAIN,
            "R1",
            "#uname eq node1",
            {},
            {},
            [reports.codes.FORCE],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.DUPLICATE_CONSTRAINTS_EXIST,
                    constraint_ids=["location-R1"],
                )
            ]
        )

    def assert_resource_in_multiinstance(
        self, resource_id, parent_type, parent_id
    ):
        self.config.runner.cib.load(
            resources=self.resources_xml, filename="cib-empty-3.9.xml"
        )
        self.env_assist.assert_raise_library_error(
            lambda: location.create_plain_with_rule(
                self.env_assist.get_env(),
                const.RESOURCE_ID_TYPE_PLAIN,
                resource_id,
                "#uname eq node1",
                {},
                {},
                [],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    force_code=reports.codes.FORCE,
                    resource_id=resource_id,
                    parent_type=parent_type,
                    parent_id=parent_id,
                )
            ]
        )

    def assert_resource_in_multiinstance_forced(
        self, resource_id, parent_type, parent_id
    ):
        constraints_xml = self.constraints_xml_minimal.format(
            resource_id=resource_id, other_constraint=""
        )
        self.config.runner.cib.load(
            resources=self.resources_xml, filename="cib-empty-3.9.xml"
        )
        self.config.env.push_cib(constraints=constraints_xml)

        location.create_plain_with_rule(
            self.env_assist.get_env(),
            const.RESOURCE_ID_TYPE_PLAIN,
            resource_id,
            "#uname eq node1",
            {},
            {},
            [reports.codes.FORCE],
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.RESOURCE_FOR_CONSTRAINT_IS_MULTIINSTANCE,
                    resource_id=resource_id,
                    parent_type=parent_type,
                    parent_id=parent_id,
                )
            ]
        )

    def test_resource_in_clone(self):
        self.assert_resource_in_multiinstance("C1R1", "clone", "C1")

    def test_resource_in_bundle(self):
        self.assert_resource_in_multiinstance("B1R1", "bundle", "B1")

    def test_resource_in_clone_forced(self):
        self.assert_resource_in_multiinstance_forced("C1R1", "clone", "C1")

    def test_resource_in_bundle_forced(self):
        self.assert_resource_in_multiinstance_forced("B1R1", "bundle", "B1")

    def test_resource_not_found(self):
        self.config.runner.cib.load(
            resources=self.resources_xml, filename="cib-empty-3.9.xml"
        )
        self.env_assist.assert_raise_library_error(
            lambda: location.create_plain_with_rule(
                self.env_assist.get_env(),
                const.RESOURCE_ID_TYPE_PLAIN,
                "RX",
                "#uname eq node1",
                {},
                {},
                [reports.codes.FORCE],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_NOT_FOUND,
                    id="RX",
                    expected_types=[],
                    context_type="",
                    context_id="",
                ),
            ]
        )

    def test_not_resource(self):
        self.config.runner.cib.load(
            resources=self.resources_xml, filename="cib-empty-3.9.xml"
        )
        self.env_assist.assert_raise_library_error(
            lambda: location.create_plain_with_rule(
                self.env_assist.get_env(),
                const.RESOURCE_ID_TYPE_PLAIN,
                "R1-monitor",
                "#uname eq node1",
                {},
                {},
                [reports.codes.FORCE],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.ID_BELONGS_TO_UNEXPECTED_TYPE,
                    id="R1-monitor",
                    expected_types=sorted(cib_const.TAG_LIST_CONSTRAINABLE),
                    current_type="op",
                ),
            ]
        )

    def test_errors(self):
        self.config.runner.cib.load(
            resources=self.resources_xml, filename="cib-empty-3.9.xml"
        )
        self.env_assist.assert_raise_library_error(
            lambda: location.create_plain_with_rule(
                self.env_assist.get_env(),
                "bad resource id type",
                "R1",
                "#uname eq",
                {
                    "bad": "option",
                    "score": "bad score",
                    "score-attribute": "some attribute",
                    "role": "bad role",
                    "id": "R1-monitor",
                },
                {
                    "bad": "option",
                    "id": "B1R1",
                    "resource-discovery": "bad discovery",
                },
                [reports.codes.FORCE],
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_ID_TYPE,
                    id_type="bad resource id type",
                    allowed_types=["resource_id_plain", "resource_id_regexp"],
                ),
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["bad"],
                    allowed=["id", "resource-discovery"],
                    option_type="constraint",
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.ID_ALREADY_EXISTS,
                    id="B1R1",
                ),
                fixture.warn(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="resource-discovery",
                    option_value="bad discovery",
                    allowed_values=[
                        CibResourceDiscovery.ALWAYS,
                        CibResourceDiscovery.EXCLUSIVE,
                        CibResourceDiscovery.NEVER,
                    ],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["bad"],
                    allowed=["id", "role", "score", "score-attribute"],
                    option_type="rule",
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.MUTUALLY_EXCLUSIVE_OPTIONS,
                    option_names=["score", "score-attribute"],
                    option_type="rule",
                ),
                fixture.error(
                    reports.codes.ID_ALREADY_EXISTS,
                    id="R1-monitor",
                ),
                fixture.error(
                    reports.codes.INVALID_SCORE,
                    score="bad score",
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="role",
                    option_value="bad role",
                    allowed_values=(
                        const.PCMK_ROLE_PROMOTED,
                        const.PCMK_ROLE_UNPROMOTED,
                    ),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.RULE_EXPRESSION_PARSE_ERROR,
                    rule_string="#uname eq",
                    reason="Expected <attribute value>",
                    rule_line="#uname eq",
                    line_number=1,
                    column_number=10,
                    position=9,
                ),
            ]
        )
